/* Copyright (C) 2003 MySQL AB

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#define BUFFER_LEN 256
#define DEFAULT_SCHEMA_NAME "DEFAULT_SCHEMA"
#define DEFAULT_CATALOG_NAME "DEFAULT_CATALOG"
#define ADMIN_DUMP_VERSION "1.0"

#define INITIAL_PACKET_LENGTH 1024*1024 /* initial size of 1M */
#define PACKET_RESIZE_INCREMENT 1024*256 /* increase in 256k increments */

//#define DEFAULT_CHARSET    "   "

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <pcre.h>
#include "myx_admin_library.h"
#include "myx_library.h"
#include <errno.h>
#include <ctype.h>
#include <glib.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <libxml/xpath.h>
#include "myx_xml_aux_functions.h"
#include <time.h>

/* structure declarations */

/* struct for get_next_sql_statement */
typedef struct {
 
  int state;
  bigint bytes_read;

} WORKING_ENVIRONMENT;

typedef struct
{
  FILE *sql_file;
  MYX_BACKUP_CONTENT *backup_content;
  WORKING_ENVIRONMENT *we;
  pcre *re_create;
  pcre *re_db_use;
  pcre_extra *pe_create;
  pcre_extra *pe_db_use;
  char *cur_db_name;
  const char *default_catalog_name;
  const char *default_schema_name;
} MYX_BCS_STATUS;

typedef struct
{
  FILE *sql_file;
  MYX_BACKUP_CONTENT *backup_content;
  WORKING_ENVIRONMENT *we;
  pcre *re_create; 
  pcre *re_db_use;
  pcre *re_drop;
  pcre *re_insert;
  pcre *re_alter;
  pcre *re_lock;
  pcre *re_unlock;
  pcre *re_set;  
  pcre *re_create_database;
  const char *target_catalog;
  const char *target_schema;
  char *current_schema_in_sqlfile; 
  char *current_schema_in_db; /* gets updated on 'USE db' */
  char *current_catalog;
  MYSQL *mysql;
  int options;
  char *buffer;
  int buffer_len;
  void (*report_warning) (const char *msg, void *user_data);
  void *report_warning_data;
} MYX_RBS_STATUS;

/* for backing up to sql-files */
typedef struct
{
  MYSQL_RES *mysql_result;
  FILE *sql_file;
  MYX_BACKUP_CONTENT *backup_content;
  MYSQL *mysql;
  int options;
  unsigned int current_table_index;
  int table_finished;
  char *current_table_quoted;
  char *current_schema_quoted;
  int current_table_rows_processed;
  int current_table_rows;
  char *extended_insert;
  int starting;
  char quote_char;
} MYX_BS_STATUS; 

/* forward declarations */
static int get_next_sql_statement(WORKING_ENVIRONMENT *we, FILE *sql_file, char **buffer, int *buffer_len, int resize_buffer,  MYX_BACKUP_ERROR *error);
static WORKING_ENVIRONMENT* new_working_environment(void);
static void free_working_environment(WORKING_ENVIRONMENT *we);
static int find_table(MYX_BACKUP_CONTENT *b, const char *table_name, const char *schema_name, const char *catalog_name);
static void free_backup_table_content(MYX_BACKUP_TABLE *t);
static char* unquote_name(char *table_name);
static int interact_with_server(MYX_RBS_STATUS *rs, char *buffer, int buffer_len, MYX_BACKUP_ERROR *error);
static int comfort_query(MYX_RBS_STATUS *rs, char *buffer, int buffer_len, MYX_BACKUP_ERROR *error);
static char *quote_name(const char *name, char *buff, char quote_char);
static int write_create_statement_to_file(MYX_BS_STATUS *status,MYX_BACKUP_ERROR *error);
static int write_row_to_file(MYX_BS_STATUS *status, MYSQL_ROW row, MYX_BACKUP_ERROR *error);
static int write_sql_file_header(MYX_BS_STATUS *status, MYX_BACKUP_ERROR *error);
static int write_sql_file_footer(MYX_BS_STATUS *status, MYX_BACKUP_ERROR *error);
static int  compare_backup_tables(const void *a, const void *b);
static int write_path_info_to_file(MYX_BS_STATUS *status, MYX_BACKUP_ERROR *error);
static MYX_BACKUP_CONTENT* copy_backup_content(MYX_BACKUP_CONTENT *content);
static int make_state_transition(int state, int c);
static int in_backup_content(const char *schema_name, MYX_BACKUP_CONTENT *backup_content);

static MYX_BACKUP_PROFILE *read_in_backup_profile(xmlNodePtr backup_node);
static void read_in_table(xmlNodePtr table_node, MYX_BACKUP_TABLE *table);
static char* check_statement(const char *line, int line_len, pcre *re, pcre_extra *pe);
static int prepare_line_is_create_statement(pcre **re, pcre_extra **pe);
static void finalize_line_is_create_statement(pcre *re, pcre_extra *pe);
static int prepare_line_is_db_use_statement(pcre **re, pcre_extra **pe);
static void finalize_line_is_db_use_statement(pcre *re, pcre_extra *pe);

MYX_BCS_STATUS* myx_new_bcs_status(const char *filename,MYX_BACKUP_ERROR *error);
int myx_free_bcs_status(MYX_BCS_STATUS *b, int free_backup_content);
int myx_get_backup_content_from_sql_file_incremental(MYX_BCS_STATUS *bc_status, int granularity, MYX_BACKUP_ERROR *error);

MYX_RBS_STATUS * myx_new_rbs_status(MYSQL *mysql, const char *filename, MYX_BACKUP_CONTENT *content, const char *target_catalog, const char *target_schema, int options, MYX_BACKUP_ERROR *error, void (*report_warning) (const char *msg, void *user_data), void *user_data);
int myx_free_rbs_status(MYX_RBS_STATUS *rs);
int myx_restore_backup_from_sql_file_incremental(MYX_RBS_STATUS *rs, int granularity, MYX_BACKUP_ERROR *error);

MYX_BS_STATUS * myx_new_bs_status(MYSQL *mysql, const char *filename, MYX_BACKUP_CONTENT *content, MYX_BACKUP_OPTIONS options, MYX_BACKUP_ERROR *error);
int myx_free_bs_status(MYX_BS_STATUS *b);
int myx_make_backup_to_sql_file_incremental(MYX_BS_STATUS *status, unsigned int num_rows, MYX_BACKUP_ERROR *error);

int myx_set_bs_options(MYX_BS_STATUS *status, int options); //TODO: remove that function?


/* taken from Viktor Vagins patch */
#define SAFE_IO(func) {func; if (errno == ENOSPC){*error=MYX_BACKUP_OUTPUTDEVICE_FULL;return -1;}}

/*
 * Public functions
 */

/* Profile functions */

/*
 *----------------------------------------------------------------------
 *
 *
 * SYNOPSIS
 *  profile_directory must include a trailing path delimiter
 * DESCRIPTION
 *
 * RETURN VALUE
 *
 * NOTES
 *----------------------------------------------------------------------
 */
MYX_ADMIN_LIB_ERROR myx_save_profile(const char *profile_name, const char *profile_directory, MYX_BACKUP_PROFILE *backup_profile)
{
  xmlDocPtr doc;
  xmlNodePtr root_node, tables_node, table_node;
  unsigned int i;
  char *filename;
  
  if ((profile_name == NULL) || (profile_directory == NULL) || (backup_profile == NULL)) return -1;

  doc = xmlNewDoc("1.0");

  root_node = doc->children = xmlNewDocRawNode(doc,NULL, "backup_profile", NULL);

  xmlNewTextChild(root_node, NULL, "profile_name", backup_profile->profile_name);
  xmlNewTextChild(root_node, NULL, "target_directory", backup_profile->target_directory);
  xmlNewTextChild(root_node, NULL, "filename_prefix", backup_profile->filename_prefix);
  xmlNewTextChild(root_node, NULL, "last_used", backup_profile->last_used);
  xmlNewTextChild(root_node, NULL, "options", int_to_str(backup_profile->options));
  xmlNewTextChild(root_node, NULL, "backup_type", int_to_str(backup_profile->backup_type));
  tables_node= xmlNewTextChild(root_node, NULL, "tables", NULL);
  
  if (backup_profile->backup_content)
  {
    for (i=0; i < backup_profile->backup_content->tables_num; i++)
    {
      table_node= xmlNewTextChild(tables_node, NULL, "table", NULL);

      xmlNewTextChild(table_node, NULL, "name", backup_profile->backup_content->tables[i].table);
      xmlNewTextChild(table_node, NULL, "schema", backup_profile->backup_content->tables[i].schema);
      xmlNewTextChild(table_node, NULL, "catalog", backup_profile->backup_content->tables[i].catalog);
    }
  }


  filename= g_strconcat(profile_directory, profile_name, NULL);
  
  if (xmlSaveFormatFile(filename, doc, 1) == -1)
  {
    xmlFreeDoc(doc);
    g_free(filename);
    return -1;
  }
  
  g_free(filename);
  xmlFreeDoc(doc);
  return 0;
}

int myx_free_profile(MYX_BACKUP_PROFILE *profile)
{
  if (profile)
  {
    xmlFree(profile->profile_name);
    xmlFree(profile->target_directory);
    xmlFree(profile->filename_prefix);
    xmlFree(profile->last_used);
    if (profile->backup_content)
      myx_free_backup_content(profile->backup_content);
  
    g_free(profile);
  }
  return 0;
}

MYX_BACKUP_PROFILE *myx_load_profile(const char *profile_name, const char *profile_directory, MYX_ADMIN_LIB_ERROR *error_code)
{
  xmlDocPtr doc;
  xmlNodePtr root;
  MYX_BACKUP_PROFILE *backup_profile;
  char *filename;
  
  *error_code = MYX_ADMIN_NO_ERROR;

  filename= g_strconcat(profile_directory, profile_name, NULL);
  
  if (!file_exists(filename))
  {
    *error_code = MYX_ERROR_CANT_OPEN_FILE;
    return NULL;
  }

  doc = xmlParseFile(filename);

  if (doc == NULL )
  {
    *error_code = MYX_XML_PARSE_ERROR;
    return NULL;
  }

  root = xmlDocGetRootElement(doc);

  if (root == NULL)
  {
    *error_code = MYX_XML_EMPTY_DOCUMENT;
    xmlFreeDoc(doc);
    return NULL;
  }

  if (xmlStrcmp(root->name, (const xmlChar *) "backup_profile"))
  {
    *error_code =  MYX_XML_NO_VALID_DOCUMENT;
    xmlFreeDoc(doc);
    return NULL;
  }

  backup_profile= read_in_backup_profile(root);

  xmlFreeDoc(doc);

  return backup_profile;
}

MYX_BACKUP_ERROR myx_make_backup_with_profile(MYSQL *mysql,  MYX_BACKUP_PROFILE *profile, const char *path, int callback_interval,
                                              int (*progress_report)(const char* current_table_name, int num_tables, int num_tables_processed, int num_rows,int num_rows_processed, void *user_data),
                                              void *user_data)
{
  MYX_BACKUP_ERROR error;

  error= myx_make_backup(mysql, path, profile->backup_content, profile->backup_type, profile->options, callback_interval, progress_report, user_data);

  return error;
}


/* MAKE-BACKUP functions */


/*
 *----------------------------------------------------------------------
 * Starts a backup of the given content. The content can consist of several tables
 * from different schemas and catalogs. display_backup_progress points to a callback
 * function to display the current progress of the backup.
 *
 * SYNOPSIS
 *   mysql : The mysql-connection that will be used
 *   filename: The name of the file that will hold the backup
 *   content: lists all tables that should be backed up
 *   backup_type: at the moment only MYX_BT_SQL_SCRIPT
 *   options: see the description
 *   progress_report: a function that will be called
 *   user_data: a pointer to user-defined data to be passed to the progress_report
 *   callback_interval: determines how many rows must be processed so that the
 *                      callback-function is called)
 * DESCRIPTION
 *  There are lot of options that determine how the actual backup is
 *  done:
 *   MYX_B_NO_CREATES:           no create Table statements will be written
 *   MYX_B_NO_EXTENDED_INSERT:   instead of mysql's faster insert syntax that
 *                                 allows several values in one insert the ansi-compatible
 *                                 normal syntax is used
 *   MYX_B_ADD_DROP_TABLE:       before each create-statement a "DROP TABLE IF EXISTS" is written
 *   MYX_B_COMMENT:              write some default comments at the beginning of the file
 *   MYX_B_DONT_WRITE_FULL_PATH  don't write "create database"/"use db"-statements to the file
 *   MYX_B_LOCK_ALL_TABLES       All tables are locked with FLush read locks before doing anything
 *   MYX_B_SINGLE_TRANSACTION    Let the backup be done in a single transaction (for innodb),
 *                                 Mutually exclusive with MYX_B_LOCK_ALL_TABLES
 *   MYX_B_DISABLE_KEYS          write  ALTER TABLE %s DISABLE KEYS before the insers; and
 *                                ALTER TABLE %s ENABLE KEY after all inserts; makes inserting
 *                               into myisam-tables a lot faster
 *   MYX_B_COMPLETE_INSERTS       INSERTS are of the form "insert into (col1,col2) values (a,b)"
 *                                instead of "insert into values (a,b)" 
 *   MYX_B_ANSI_QUOTES=512        Identifiers are quoted with " instead of mysql's normal `
 *   MYX_B_ADD_LOCKS              Adds a LOCK TABLES statement before an insert-block and 
 *                                an UNLOCK TABLES afterwards
 ** RETURN VALUE
 *  0 on success, non-zero to indicate an error
 * NOTES
 *  if a file with filename already exists, it will be overwritten.
 *
 *----------------------------------------------------------------------
 */
MYX_BACKUP_ERROR myx_make_backup(MYSQL *mysql, const char *filename, MYX_BACKUP_CONTENT *content, MYX_BACKUP_TYPE backup_type,  int options, int callback_interval,
                                 int (*progress_report)(const char* current_table_name, int num_tables, int num_tables_processed, int num_rows,int num_rows_processed,
							void *user_data),
                                 void *user_data)
{
  if (backup_type == MYX_BT_SQL_SCRIPT)
  {
    MYX_BS_STATUS *bs_status;
    char *last_table= "";
    int x= 0;
    MYX_BACKUP_ERROR error=0;

    if (! (bs_status= myx_new_bs_status(mysql, filename, content, options, &error)) )
      return error;

    x= callback_interval;

    /* backup one line at a time */
    while ( myx_make_backup_to_sql_file_incremental(bs_status, 1, &error) > 0)
    {
      int new_table= strcmp(last_table, bs_status->current_table_quoted);

      x--;
      if ((progress_report) && (new_table || x <= 0))
      {
        if (new_table)
        {
          last_table= bs_status->current_table_quoted;
        }
        x= callback_interval;

        if ( (*progress_report)(bs_status->current_table_quoted,bs_status->backup_content->tables_num , 
				bs_status->current_table_index+1, bs_status->current_table_rows,
				bs_status->current_table_rows_processed, user_data) )
        {
          myx_free_bs_status(bs_status);
          return -1;
        }
      }
    }
    /* end while */
    myx_free_bs_status(bs_status);
    return error;
  }

  return -1; /* unknown backup type */
}

/*
 *----------------------------------------------------------------------
 * Creates the MYX_BS_STATUS, that saves needed information for
 * myx_make_backup_to_sql_file_incremental(MYX_BS_STATUS *status, int num_rows).
 *
 * SYNOPSIS
 *   content: specifies which tables are to be backed up
 *   options: 0 or a combination of options;
 *            if 0 the default-options will be used
 * DESCRIPTION
 *
 * RETURN VALUE
 *
 * NOTES
 *----------------------------------------------------------------------
 */
MYX_BS_STATUS *myx_new_bs_status(MYSQL *mysql, const char *filename, MYX_BACKUP_CONTENT *content, MYX_BACKUP_OPTIONS options, MYX_BACKUP_ERROR *error)
{
  MYX_BS_STATUS *b;

  if (options & MYX_B_LOCK_ALL_TABLES & MYX_B_SINGLE_TRANSACTION)
  {
    *error= MYX_BACKUP_ILLEGAL_OPTION;
    return NULL;
  }
  b= g_malloc0(sizeof(MYX_BS_STATUS) );

  if (! (b->sql_file=fopen(filename, "w")) )
  {
    *error=  MYX_BACKUP_CANT_OPEN_FILE;
    g_free(b);
    return NULL;
  }

  b->mysql= mysql;
  b->backup_content= copy_backup_content(content);
  b->options= options;
  b->current_table_index= -1;
  b->starting= 1; /* set to display that we have not fetched anything yet */

  qsort(b->backup_content->tables, b->backup_content->tables_num, sizeof(MYX_BACKUP_TABLE), compare_backup_tables);

  if (b->options == 0)
  {
    b->options |=  MYX_B_ADD_DROP_TABLE;
    b->options |=  MYX_B_COMMENT;
    b->options |=  MYX_B_DISABLE_KEYS;
    b->options |=  MYX_B_ADD_LOCKS;
  }

  if (b->options & MYX_B_ANSI_QUOTES)
    b->quote_char= '"';
  else
    b->quote_char= '`';

  return b;
}

/*
 *----------------------------------------------------------------------
 * Overwrites the current options for a backup-operation with the
 * options specified in the argument
 *
 * SYNOPSIS
 *
 * DESCRIPTION
 *  Should be called only right after myx_new_bs_status.
 *
 * RETURN VALUE
 *  If the options argument is negative, then the current options
 *  value of status is returned.
 *  The newly set option are returned otherwise.
 * NOTES
 *----------------------------------------------------------------------
 */
int myx_set_bs_options(MYX_BS_STATUS *status, int options)
{
  if (options < 0)
    return status->options;

  return (status->options= options);
}

int myx_free_bs_status(MYX_BS_STATUS *b)
{
  if (b)
  {
    if (b->sql_file)
      fclose(b->sql_file);

    if (b->current_table_quoted)
      g_free(b->current_table_quoted);
    if (b->current_schema_quoted)
      g_free(b->current_schema_quoted);
    if (b->extended_insert)
      g_free(b->extended_insert);
    myx_free_backup_content(b->backup_content);

    g_free(b);

  }
  return 0;
}



/*
 *----------------------------------------------------------------------
 * backs up num-rows from the current table
 *
 * SYNOPSIS
 *
 * DESCRIPTION
 *
 * RETURN VALUE
 *  a value greater 0 if we should continue
 *  0 or negative in case an error has happened or if we are finished
 * NOTES
 *----------------------------------------------------------------------
 */
int myx_make_backup_to_sql_file_incremental(MYX_BS_STATUS *status, unsigned int num_rows, MYX_BACKUP_ERROR *error)
{
  unsigned int count=0;
  MYSQL_ROW row;

  /* if we don't have any more tables to back up.. */
  if ( ((status->table_finished) && ((status->current_table_index+1)) >= status->backup_content->tables_num))
  {
    if (write_sql_file_footer(status, error))
    {
      return -1;
    }

    return 0;
  }

  /* if we start fetching rows of a new table */
  if (status->table_finished || status->starting)
  {
    gchar *sel_stmt;
    MYSQL_RES *result;
    char *current_table, *current_schema;

    if (status->starting)
    {
      if (write_sql_file_header(status, error))
	return -1;

      status->starting=0;
      if (status->options & MYX_B_LOCK_ALL_TABLES)
      {
        if (myx_mysql_query(status->mysql, "FLUSH TABLES WITH READ LOCK"))
        {
          *error= MYX_BACKUP_MYSQL_ERROR;
          return -1; /* fatal */
        }
      }
      else if (status->options & MYX_B_SINGLE_TRANSACTION)
      {
        if (myx_mysql_query(status->mysql, "BEGIN") )
        {
          *error= MYX_BACKUP_MYSQL_ERROR;
          return -1;  /* fatal */
        }
      }
      if (status->options & MYX_B_ANSI_QUOTES)
      {
	if (myx_mysql_query(status->mysql, "SET @OLD_SQL_MODE=@@SQL_MODE,sql_mode='ANSI_QUOTES'"))
	{
          *error= MYX_BACKUP_MYSQL_ERROR;
	  return -1;
	}
      }
    }
    /* end if (status->starting) */

    status->table_finished= 0;
    status->current_table_index++;
    current_table= status->backup_content->tables[status->current_table_index].table;
    current_schema= status->backup_content->tables[status->current_table_index].schema;

    if (status->current_table_quoted)
      g_free(status->current_table_quoted);
    status->current_table_quoted= g_malloc((unsigned int)strlen(current_table)*2+3);
    quote_name(current_table, status->current_table_quoted, status->quote_char);

    if (status->current_schema_quoted)
      g_free(status->current_schema_quoted);
    status->current_schema_quoted= g_malloc((unsigned int) strlen(current_schema)*2 +3);
    quote_name(current_schema, status->current_schema_quoted, status->quote_char);

    status->current_table_rows_processed= 0;

    /* find out how many rows this table has.. */

    sel_stmt= g_strconcat("SELECT count(*) FROM ", status->current_schema_quoted, ".", status->current_table_quoted, NULL);
    if ( myx_mysql_query(status->mysql, sel_stmt) || !(result= mysql_store_result(status->mysql)) )
    {
      g_free(sel_stmt);
      *error =  MYX_BACKUP_MYSQL_ERROR;
      return 0;
    }
    g_free(sel_stmt);

    row= mysql_fetch_row(result);
    if (row) status->current_table_rows= atoi(row[0]);
    else
    {
      mysql_free_result(result);
      *error =  MYX_BACKUP_MYSQL_ERROR;
      return 0;
    }
    mysql_free_result(result);

    /* write the path-information for this table */
    if (!(status->options & MYX_B_DONT_WRITE_FULL_PATH))
    {
      if (write_path_info_to_file(status, error))
      {
	return -1;
      }
    }

    /* get the create-statement of the new table */
    if (! (status->options & MYX_B_NO_CREATES) )
    {
      if (write_create_statement_to_file(status, error))
      {
	return -1;
      }
    }
    /* get the rows of the new table */
    sel_stmt= g_strconcat("SELECT /*!40001 SQL_NO_CACHE */ * FROM ", status->current_schema_quoted, ".", status->current_table_quoted, NULL);
    if (myx_mysql_query(status->mysql, sel_stmt) || !(status->mysql_result= mysql_use_result(status->mysql)) )
    {
      g_free(sel_stmt);
      *error =  MYX_BACKUP_MYSQL_ERROR;
      return 0;
    }
    g_free(sel_stmt);

    if (status->options &  MYX_B_DISABLE_KEYS)
      SAFE_IO(fprintf(status->sql_file, "\n/*!40000 ALTER TABLE %s DISABLE KEYS */;\n", status->current_table_quoted));

    if (status->options & MYX_B_ADD_LOCKS)
      SAFE_IO(fprintf(status->sql_file,"LOCK TABLES %s WRITE;\n", status->current_table_quoted));
  }

  /* we will fetch num_rows rows every time (but less if the table has no more rows */
  while ((count < num_rows)  && (row= mysql_fetch_row(status->mysql_result)) )
  {
    if (write_row_to_file(status, row, error))
	return -1;

    status->current_table_rows_processed++;
    count++;
  }

  /* if backing up this table has ended */
  if (status->current_table_rows_processed == status->current_table_rows)
  {
    if (row)
    {
      row= mysql_fetch_row(status->mysql_result);  /* just to be sure because the documentation says that you have to call fetch_row until it returns NULL.. */
    }
    assert(row == NULL);

    status->table_finished= 1;
    mysql_free_result(status->mysql_result); /* free the result of the last select * query */
    status->mysql_result= NULL;

    /* it is possible that our extended insert needs to be written out */
    if ( !(status->options & MYX_B_NO_EXTENDED_INSERT) && (status->extended_insert) )
    {
      unsigned int slen= (unsigned int)strlen(status->extended_insert);
      status->extended_insert[slen-1]= ';';  /*overwrite the comma with ;*/
      SAFE_IO(fprintf(status->sql_file, "%s\n",status->extended_insert));
      g_free(status->extended_insert);
      status->extended_insert= NULL;
    }
    if (status->options & MYX_B_ADD_LOCKS)
      SAFE_IO(fprintf(status->sql_file,"UNLOCK TABLES;\n"));

    if (status->options &  MYX_B_DISABLE_KEYS)
      SAFE_IO(fprintf(status->sql_file, "\n/*!40000 ALTER TABLE %s ENABLE KEYS */;\n", status->current_table_quoted));

    /* if this is the last table */
    if (status->current_table_index+1 >= status->backup_content->tables_num )
    {
      if (status->options & MYX_B_LOCK_ALL_TABLES)
      {
	if (myx_mysql_query(status->mysql, "UNLOCK TABLES"))
	{
	  *error =  MYX_BACKUP_MYSQL_ERROR;
	  return -1;
	}
      }
      else if (status->options & MYX_B_SINGLE_TRANSACTION)
      {
        if (myx_mysql_query(status->mysql, "COMMIT") )
        {
          *error= MYX_BACKUP_MYSQL_ERROR;
          return -1;  /* fatal */
        }
      }

      if (status->options & MYX_B_ANSI_QUOTES)
      {
        if (myx_mysql_query(status->mysql, "SET sql_mode=@OLD_SQL_MODE"))
        {
	        *error= MYX_BACKUP_MYSQL_ERROR;
	        return -1;
        }
      }
    }
  }

  return 1;
}

/* GET-BACKUP-CONTENT Functions */

/*
 *----------------------------------------------------------------------
 * Returns all catalogs/schemas/tables stored in the backup
 *
 * SYNOPSIS
 *   default_catalog_name: If no catalog is given for a table in
 *                  the given file, it is assumed to be in the
 *                  default_catalog_name
 *   default_schema_name: if a table in the given file has no
 *                  schema(=database) then it is assumed that
 *                  the table lies in the default_schema_name
 *   report_interval: report callback will be called every this many
 *                  bytes of data is processed
 *   progress_report: call-back function
 *                    returns 0 if we should continue our work
 *                    progress_report may be NULL
 *   user_data:     user specified data to be passed back when
 *                  progress_report is called
 * DESCRIPTION
 *
 * RETURN VALUE
 *
 * NOTES
 *
 *----------------------------------------------------------------------
 */
MYX_BACKUP_CONTENT *myx_get_backup_content(const char *filename,  MYX_BACKUP_TYPE backup_type, int report_interval, int (*progress_report) (bigint bytes_read, bigint bytes_total, void *user_data), void *user_data, MYX_BACKUP_ERROR *error )
{
  MYX_BACKUP_CONTENT *content;
  if (backup_type == MYX_BT_SQL_SCRIPT)
  {
    MYX_BCS_STATUS *stat;
    bigint next_call, file_size;

    *error=0;

    if (! (stat= myx_new_bcs_status(filename, error)) )
      return NULL;

    file_size= get_file_size(filename);
    next_call= report_interval;

    while (myx_get_backup_content_from_sql_file_incremental(stat, 1, error) > 0 )
    {
      if ( (progress_report) && (stat->we->bytes_read >= next_call) )
      {
        next_call += report_interval;
        if ((*progress_report)(stat->we->bytes_read, file_size, user_data))
        {
          myx_free_bcs_status(stat,1);
          return NULL;
        }
      }
    }

    if (*error) 
    {
      myx_free_bcs_status(stat, 1);
      return NULL;
    }
    
    if (progress_report && (file_size % report_interval))
      (*progress_report)(stat->we->bytes_read, file_size, user_data);

    content= stat->backup_content;
    myx_free_bcs_status(stat,0);
    return content;
  }

  return NULL;
}

int myx_free_backup_content(MYX_BACKUP_CONTENT *backup)
{
  unsigned int i;

  if (backup)
  {
    for (i=0; i< backup->tables_num; i++)
    {
      free_backup_table_content(backup->tables+i);
    }
    g_free(backup->tables);
    g_free(backup);
  }

  return 0;
}

MYX_BCS_STATUS* myx_new_bcs_status(const char *filename, MYX_BACKUP_ERROR *error)
{
  MYX_BCS_STATUS *b;

  b= g_malloc0( sizeof(MYX_BCS_STATUS) );

  if (! (b->sql_file=fopen(filename, "r")) )
  {
    *error= MYX_BACKUP_CANT_OPEN_FILE;
    g_free(b);
    return NULL;
  }

  /*
  if (setvbuf(b->sql_file, b->file_buffer, _IOFBF, sizeof(b->file_buffer)) )
  {
    *error= MYX_BACKUP_CANT_OPEN_FILE;
    g_free(b);
    return NULL;
  } */

  b->backup_content= g_malloc0(sizeof(MYX_BACKUP_CONTENT) );
  b->we= new_working_environment();

  prepare_line_is_create_statement(&b->re_create, &b->pe_create);
  prepare_line_is_db_use_statement(&b->re_db_use, &b->pe_db_use);

  b->default_schema_name= DEFAULT_SCHEMA_NAME;
  b->default_catalog_name= DEFAULT_CATALOG_NAME;

  return b;
}

int myx_free_bcs_status(MYX_BCS_STATUS *b, int free_backup_content)
{
  if (b)
  {
    fclose(b->sql_file);
    free_working_environment(b->we);
    g_free(b->cur_db_name);
    if (free_backup_content) g_free(b->backup_content);
    finalize_line_is_db_use_statement(b->re_db_use, b->pe_db_use);
    finalize_line_is_create_statement(b->re_create, b->pe_create);

    g_free(b);
  }

  return 0;
}

/*
 *----------------------------------------------------------------------
 *
 *
 * SYNOPSIS
 *
 * DESCRIPTION
 *  granularity = number of sql-statements
 * RETURN VALUE
 *  the number of successfully processed sql-statements
 *  0 if there are no more sql-statements
 *  a negative number in case of an error
 * NOTES
 *----------------------------------------------------------------------
 */
int myx_get_backup_content_from_sql_file_incremental(MYX_BCS_STATUS *bc_status, int granularity, MYX_BACKUP_ERROR *error)
{
  int len, tables_found;
  char buffer[BUFFER_LEN];
  char *bufptr= (char*)buffer;
  int buffer_len= sizeof(buffer);
  int count= 0;
  MYX_BACKUP_TABLE *bt;
  char *tmp;

  while( (count < granularity) &&  (len= get_next_sql_statement(bc_status->we, bc_status->sql_file, &bufptr, &buffer_len, 0, error)) )
  {
    /* we have to 0 the quote flag here because it's a leftover from a line that was only partially read.
     * since we're only interested in the beginning of the lines here, we
     * will ignore the leftovers from the previous line -alfredo */
    //bc_status->we->quote= 0;
    
    if (*error)
      return -1;
    
    /* do some checks now */

    if ( (tmp=check_statement(buffer,len,bc_status->re_create,bc_status->pe_create)) ) /* CREATE TABLE */
    {
      tables_found= ++bc_status->backup_content->tables_num;
      bc_status->backup_content->tables= g_realloc(bc_status->backup_content->tables, tables_found * sizeof(MYX_BACKUP_TABLE));

      bt= bc_status->backup_content->tables + tables_found-1;;
      bt->catalog= g_strdup(bc_status->default_catalog_name);
      bt->schema= (bc_status->cur_db_name) ? g_strdup(bc_status->cur_db_name) : g_strdup(bc_status->default_schema_name);
      //bt->schema= unquote_name(bt->schema);
      bt->table= unquote_name(tmp);
    }
    else if ( (tmp=check_statement(buffer,len,bc_status->re_db_use,bc_status->pe_db_use)) ) /* USE */
    {
      if (bc_status->cur_db_name)
        g_free(bc_status->cur_db_name);

      bc_status->cur_db_name= unquote_name(tmp);

    }

    count++;
  }

  return count;
}

/* RESTORE-BACKUP Functions */

/*
 *----------------------------------------------------------------------
 * Restores a backup with the given filename into the database that
 * is reached by the mysql-struct, but only the catalogs/schemas/tables
 * specified by content. The backup can be restored in a different catalog and
 * or schema.
 *
 * SYNOPSIS
 *   target_catalog: the catalog that the table should be restored in;
 *                   if NULL the catalog in content will be used
 *   target_schema:  the schema that the table should be restored in;
 *                   if NULL the schema in content will be used
 *   content:        specifies with tables (identified by tablename,schema-
 *                   name and catalogname) that are defined in filename
 *                   are to be restored
 *   progress_report,
 *   report_interval: See myx_get_backup_content
 *   report_warning: this function will be called to report a not-fatal
 *                   error. If NULL the warning is printed to stderr
 * DESCRIPTION
 *  Possible option values:
 *    -) MYX_RBS_FORCE  continue even in case of an sql-error
 *    -) MYX_RBS_DONT_CREATE_TARGETS   Don't create the target schema and/or target catalog if it doesn't exist
 * RETURN VALUE
 *  0 on success, errorcode otherwise
 * NOTES
 *----------------------------------------------------------------------
 */
MYX_BACKUP_ERROR myx_restore_backup(MYSQL *mysql, const char *filename, MYX_BACKUP_CONTENT *content, const char *target_catalog, const char *target_schema,
                                    MYX_BACKUP_TYPE backup_type, int options, int report_interval,  int (*progress_report) (bigint bytes_read, bigint bytes_total, void *user_data), void *ruser_data,
                                    void (*report_warning) (const char *msg, void *user_data), void *wuser_data)
{
  MYX_BACKUP_ERROR error= 0;

  if (backup_type == MYX_BT_SQL_SCRIPT)
  {
    MYX_RBS_STATUS *rs;
    bigint next_call;
    bigint file_size;

    file_size= get_file_size(filename);
    next_call= report_interval;

    if ( !(rs= myx_new_rbs_status(mysql, filename, content, target_catalog, target_schema, options, &error, report_warning, wuser_data)) )
      return error;

    while ( myx_restore_backup_from_sql_file_incremental(rs, 1, &error) > 0)
    {
      if ((progress_report) && (rs->we->bytes_read >= next_call))
      {
        next_call+= report_interval;
        if ((*progress_report)(rs->we->bytes_read, file_size, ruser_data))
        {
          myx_free_rbs_status(rs);
          return 0;
        }
      }
    }

    if ((file_size % report_interval) && (progress_report))
      (*progress_report)(rs->we->bytes_read, file_size, ruser_data);

    myx_free_rbs_status(rs);

    return error;
  }

  return 1;
}

MYX_RBS_STATUS *myx_new_rbs_status(MYSQL *mysql, const char *filename, MYX_BACKUP_CONTENT *content, const char *target_catalog,const char *target_schema,
                                   int options, MYX_BACKUP_ERROR *error, void (*report_warning) (const char *msg, void *user_data), void *user_data )
{
  MYX_RBS_STATUS *rs;
  const char *error_str;
  int erroffset;

  rs= g_malloc0( sizeof(MYX_RBS_STATUS) );

  if (! (rs->sql_file=fopen(filename, "r")) )
  {
    *error= MYX_BACKUP_CANT_OPEN_FILE;
    g_free(rs);
    return NULL;
  }

  if (! (rs->buffer= g_try_malloc(INITIAL_PACKET_LENGTH)) )
  {
    *error= MYX_BACKUP_MALLOC_FAILED;
    g_free(rs);
    return NULL;
  }
  rs->buffer_len= INITIAL_PACKET_LENGTH;

  rs->we= new_working_environment();

  prepare_line_is_create_statement(&rs->re_create, NULL);
  prepare_line_is_db_use_statement(&rs->re_db_use, NULL);

  rs->mysql= mysql;
  if(target_catalog)
  {
    if(target_catalog[0])
    {
      rs->target_catalog= target_catalog;
    }
    else
    {
      rs->target_catalog= NULL;
    }
  }
  else
  {
    rs->target_catalog= NULL;
  }
  if(target_schema)
  {
    if(target_schema[0])
    {
      rs->target_schema= target_schema;
    }
    else
    {
      rs->target_schema= NULL;
    }
  }
  else
  {
    rs->target_schema= NULL;
  }
  rs->backup_content= content;
  rs->current_schema_in_sqlfile= g_strdup(DEFAULT_SCHEMA_NAME);
  rs->current_catalog= g_strdup(DEFAULT_CATALOG_NAME);
  rs->current_schema_in_db= g_strdup(" ");
  rs->options= options;
  rs->report_warning= report_warning;
  rs->report_warning_data= user_data;

  /* setup the regular expressions */
  /* drop table may only contain one table at the moment */
  rs->re_drop= pcre_compile("drop\\s+table\\s+(?:if\\s+exists\\s+)?(\\S+)\\s*;",  PCRE_ANCHORED|PCRE_CASELESS|PCRE_DOTALL, &error_str, &erroffset, NULL);
  rs->re_insert= pcre_compile("insert\\s+into\\s+(\\S+)", PCRE_ANCHORED|PCRE_CASELESS, &error_str, &erroffset, NULL);
  rs->re_alter= pcre_compile("(?:\\/\\*\\!\\d+\\s+)?alter\\s+table\\s+(\\S+)", PCRE_ANCHORED|PCRE_CASELESS, &error_str, &erroffset, NULL);
  rs->re_lock= pcre_compile("lock\\s+tables\\s+(\\S+)\\s+write\\s*;",  PCRE_ANCHORED|PCRE_CASELESS, &error_str, &erroffset, NULL);
  /* rs->re_lock= pcre_compile("lock\\s+tables\\s+(\\S+)(?:\,\\S+\s+(?:as\\s+\S+\s+)?(read(\s+local)?|(?:low_priority\s+)?write))*",  
     PCRE_ANCHORED|PCRE_CASELESS, &error, &erroffset, NULL); */
  rs->re_unlock= pcre_compile("unlock\\s+(tables)",  PCRE_ANCHORED|PCRE_CASELESS, &error_str, &erroffset, NULL);
  rs->re_set= pcre_compile("(?:\\/\\*\\!\\d+\\s+)?(SET)",  PCRE_ANCHORED|PCRE_CASELESS, &error_str, &erroffset, NULL);
  /*rs->re_create_database= pcre_compile("create\\s+database\\s+(?:(?:if\\s+not\\s+exists\\s+)||(?:\\/\\*\\!\\d+\\s*if\\s+not\\s+exists\\*\\/\\s+))?(\\S+).*;", 
    PCRE_ANCHORED|PCRE_CASELESS|PCRE_DOTALL, &error_str, &erroffset, NULL);*/
  rs->re_create_database= pcre_compile("create\\s+database\\s+(?:\\/\\*\\!\\d+\\s*if\\s+not\\s+exists\\*\\/\\s+)?(\\S+).*;", 
				       PCRE_ANCHORED|PCRE_CASELESS|PCRE_DOTALL, &error_str, &erroffset, NULL);

  if (!rs->re_drop || !rs->re_insert || !rs->re_alter || !rs->re_lock || !rs->re_unlock || !rs->re_set || !rs->re_create_database)
  {
    free_working_environment(rs->we);
    fclose(rs->sql_file);
    finalize_line_is_create_statement(rs->re_create, NULL);
    finalize_line_is_db_use_statement(rs->re_db_use, NULL);
    g_free(rs);
    *error= MYX_BACKUP_PCRE_ERROR;
    return NULL;
  }

  return rs;
}

int myx_free_rbs_status(MYX_RBS_STATUS *rs)
{
  if (rs)
  {
    fclose(rs->sql_file);
    free_working_environment(rs->we);
    finalize_line_is_create_statement(rs->re_create, NULL);
    finalize_line_is_db_use_statement(rs->re_db_use, NULL);
    g_free(rs->current_schema_in_sqlfile);
    g_free(rs->current_catalog);

    pcre_free(rs->re_drop);
    pcre_free(rs->re_insert);
    pcre_free(rs->re_alter);
    pcre_free(rs->re_lock);
    pcre_free(rs->re_unlock);
    pcre_free(rs->re_set);
    pcre_free(rs->re_create_database);

    g_free(rs->buffer);
    g_free(rs);
  }
  return 0;
}

/* returns the number of successfully processed sql-statements
 * 0 if there is no more sql-statement
 * a negative number in case of an error
 */
int myx_restore_backup_from_sql_file_incremental(MYX_RBS_STATUS *rs, int granularity, MYX_BACKUP_ERROR *error)
{
  int len;
  int count= 0;
  char *tmp;
  char *warning;

  /* We will only execute statements we know. Otherwise we could do something bad without knowing it.  */

  while( (count < granularity) &&  (len= get_next_sql_statement(rs->we, rs->sql_file, &rs->buffer, &rs->buffer_len, 1, error)) )
  {
    if (*error)
      return -1;
    
    /* do some checks now */

    if ( (tmp=check_statement(rs->buffer,len, rs->re_insert, NULL)) ) /* INSERT */
    {
      tmp= unquote_name(tmp);

      if ( (find_table(rs->backup_content, tmp, rs->current_schema_in_sqlfile, rs->current_catalog)) )
      {
        if (interact_with_server(rs, rs->buffer, len, error)) return -1;
      }
      g_free(tmp);
    }
    else if ( (tmp=check_statement(rs->buffer,len,rs->re_drop, NULL)) ) /* DROP */
    {
      tmp= unquote_name(tmp);

      if (find_table(rs->backup_content, tmp, rs->current_schema_in_sqlfile, rs->current_catalog) )
      {
        if (interact_with_server(rs, rs->buffer, len, error)) return -1;
      }
      g_free(tmp);
    }
    else if ( (tmp=check_statement(rs->buffer,len,rs->re_alter, NULL)) ) /* ALTER */
    {
      tmp= unquote_name(tmp);

      if (find_table(rs->backup_content, tmp, rs->current_schema_in_sqlfile, rs->current_catalog) )
      {
        if (interact_with_server(rs, rs->buffer, len, error)) return -1;
      }
      g_free(tmp);
    }
    else if ( (tmp=check_statement(rs->buffer,len,rs->re_lock, NULL)) ) /* LOCK */
    {
      tmp= unquote_name(tmp);

      if (find_table(rs->backup_content, tmp, rs->current_schema_in_sqlfile, rs->current_catalog) )
      {
        if (interact_with_server(rs, rs->buffer, len, error)) return -1;
      }
      g_free(tmp);
    }
    else if ( (tmp=check_statement(rs->buffer,len,rs->re_unlock, NULL)) ) /* UNLOCK */
    {
      if (interact_with_server(rs, rs->buffer, len, error)) return -1;
      g_free(tmp);
    }
    else if ( (tmp=check_statement(rs->buffer,len,rs->re_set, NULL)) ) /* SET */
    {
      g_free(tmp);
      if (comfort_query(rs, rs->buffer, len, error)) return -1;
    }
    else if ( (tmp=check_statement(rs->buffer,len,rs->re_create, NULL)) ) /* CREATE TABLE */
    {
      tmp= unquote_name(tmp);

      if (find_table(rs->backup_content, tmp, rs->current_schema_in_sqlfile, rs->current_catalog) )
      {
        if (interact_with_server(rs, rs->buffer, len, error)) return -1;
      }
      g_free(tmp);
    }
    else if ( (tmp=check_statement(rs->buffer,len,rs->re_db_use, NULL)) ) /* USE */
    {
      tmp= unquote_name(tmp);

      if (rs->current_schema_in_sqlfile) g_free(rs->current_schema_in_sqlfile);
      rs->current_schema_in_sqlfile= tmp;

      if (!rs->target_schema) 
      {
        /* take the information from the sql-file */

	if (in_backup_content(tmp, rs->backup_content) )
	{
	  /* there is at least one table that should get restored in this schema*/

	  if (comfort_query(rs, rs->buffer, len, error)) return -1;
	}
      }
    }
    else if ( (tmp=check_statement(rs->buffer,len,rs->re_create_database, NULL)) ) /* CREATE DATABASE */
    {
      unquote_name(tmp);
      
      if (!rs->target_schema) 
      {
        /* take the information from the sql-file */

	if (in_backup_content(tmp, rs->backup_content) )
	{
	  /* there is at least one table that should get restored in this schema*/

	  if (comfort_query(rs, rs->buffer, len, error)) return -1;
	}
      }
      g_free(tmp); 
    }
    else
    {
      /*Unknown statement*/
      warning= g_strconcat("Warning: Do not know how to handle this statement:\n", rs->buffer,
                           "\nIgnoring this statement. Please file a bug-report including the statement if this statement should be recognized.\n",NULL);
      if (rs->report_warning)
        (*rs->report_warning)(warning, rs->report_warning_data);
      else
        fprintf(stderr,"%s", warning);

      g_free(warning);
    }
    count++;
  }

  return count;
}



/*
 * ******************
 *
 * Private functions
 *
 * ******************
 */


/* returns true if backup_content contains a table in schema schema_name */
static int in_backup_content(const char *schema_name, MYX_BACKUP_CONTENT *backup_content)
{
  unsigned int i;

  for (i=0; i < backup_content->tables_num; i++)
  {
    if ((backup_content->tables[i].schema) && (!strcmp(backup_content->tables[i].schema, schema_name)) )
      return 1;
  }

  return 0;
}


static MYX_BACKUP_PROFILE *read_in_backup_profile(xmlNodePtr backup_node)
{
  MYX_BACKUP_PROFILE *backup_profile;
  xmlNodePtr cur;
  xmlDocPtr doc;
  
  doc= backup_node->doc;
  
  backup_profile= g_malloc0(sizeof(MYX_BACKUP_PROFILE));
  
  backup_profile->backup_content= g_malloc0(sizeof(MYX_BACKUP_CONTENT));

  cur = backup_node->children;
  while (cur)
  {
    if ( !xmlStrcmp(cur->name, "profile_name") &&  cur->type == XML_ELEMENT_NODE)
    {
      backup_profile->profile_name= xmlNodeListGetString(doc,cur->children,1);
    }
    else if ( !xmlStrcmp(cur->name, "target_directory") &&  cur->type == XML_ELEMENT_NODE)
    {
      backup_profile->target_directory= xmlNodeListGetString(doc,cur->children,1);
    }
    else if ( !xmlStrcmp(cur->name, "filename_prefix") &&  cur->type == XML_ELEMENT_NODE)
    {
      backup_profile->filename_prefix= xmlNodeListGetString(doc,cur->children,1);
    }
    else if ( !xmlStrcmp(cur->name, "last_used") &&  cur->type == XML_ELEMENT_NODE)
    {
      backup_profile->last_used= xmlNodeListGetString(doc,cur->children,1);
    }
    else if ( !xmlStrcmp(cur->name, "options") &&  cur->type == XML_ELEMENT_NODE)
    {
      backup_profile->options= atoi_and_free(xmlNodeListGetString(doc,cur->children,1));
    }
    else if ( !xmlStrcmp(cur->name, "last_used") &&  cur->type == XML_ELEMENT_NODE)
    {
      backup_profile->last_used= xmlNodeListGetString(doc,cur->children,1);
    }
    else if ( !xmlStrcmp(cur->name, "backup_type") &&  cur->type == XML_ELEMENT_NODE)
    {
      backup_profile->backup_type= atoi_and_free(xmlNodeListGetString(doc,cur->children,1));
    }
    else if ( !xmlStrcmp(cur->name, "tables") &&  cur->type == XML_ELEMENT_NODE)
    {
      int i;
      xmlNodePtr cur2;

      backup_profile->backup_content->tables_num= get_child_count(cur, "table");
      backup_profile->backup_content->tables= g_malloc0(sizeof(MYX_BACKUP_TABLE) * backup_profile->backup_content->tables_num);

      i=0;
      cur2= cur->children;
      while (cur2)
      {
        if (!xmlStrcmp(cur2->name, "table") )
        {
          read_in_table(cur2, backup_profile->backup_content->tables+i);
          
          i++;
        }

        cur2= cur2->next;
      }
    }

    cur = cur->next;
  }
  
  return backup_profile;
}

static void read_in_table(xmlNodePtr table_node, MYX_BACKUP_TABLE *table)
{
  xmlNodePtr cur;
  xmlDocPtr doc;
  
  doc= table_node->doc;
  
  cur= table_node->children;
  while(cur)
  {
    if ( !xmlStrcmp(cur->name, "name") &&  cur->type == XML_ELEMENT_NODE)
    {
      table->table= xmlNodeListGetString(doc,cur->children,1);
    }
    else if ( !xmlStrcmp(cur->name, "schema") &&  cur->type == XML_ELEMENT_NODE)
    {
      table->schema= xmlNodeListGetString(doc,cur->children,1);
    }
    else if ( !xmlStrcmp(cur->name, "catalog") &&  cur->type == XML_ELEMENT_NODE)
    {
      table->catalog= xmlNodeListGetString(doc,cur->children,1);
    }

    cur= cur->next;
  }
}

static MYX_BACKUP_CONTENT* copy_backup_content(MYX_BACKUP_CONTENT *content)
{
  unsigned int i;
  MYX_BACKUP_CONTENT *copy;

  copy= g_malloc0(sizeof(MYX_BACKUP_CONTENT));
  copy->tables_num= content->tables_num;
  copy->tables= g_malloc0(sizeof(MYX_BACKUP_TABLE) * copy->tables_num);

  for (i=0; i< content->tables_num; i++)
  {
    copy->tables[i].catalog= g_strdup(content->tables[i].catalog);
    copy->tables[i].schema= g_strdup(content->tables[i].schema);
    copy->tables[i].table= g_strdup(content->tables[i].table);
  }

  return copy;
}

static int  compare_backup_tables(const void *a, const void *b)
{
  MYX_BACKUP_TABLE *t1, *t2;
  int cat_compare;

  t1= (MYX_BACKUP_TABLE *)a;
  t2= (MYX_BACKUP_TABLE *)b;

  /* allow NULL values for the catalogs for now */
  if ( !t1->catalog || !t2->catalog || !(cat_compare= strcmp(t1->catalog, t2->catalog)) )
  {
    return (strcmp(t1->schema, t2->schema));
  }
  else
    return cat_compare;
}

static int write_path_info_to_file(MYX_BS_STATUS *status, MYX_BACKUP_ERROR *error)
{
  MYX_BACKUP_TABLE *current, *last;
  char *database_name;

  current= status->backup_content->tables + status->current_table_index;
  last= status->backup_content->tables + status->current_table_index-1;

  /* We ignore the catalog for now */
  if ( (status->current_table_index == 0) || strcmp(current->schema, last->schema) )
  {
    MYSQL_ROW row;
    MYSQL_RES *result;
    char *stmt= g_strconcat("SHOW CREATE DATABASE WITH IF NOT EXISTS ", status->current_schema_quoted,  NULL);
    
    if ((myx_mysql_query(status->mysql, stmt)) || !(result= mysql_store_result(status->mysql)) )
    {
      SAFE_IO(fprintf(status->sql_file, "CREATE DATABASE /*!32312 IF NOT EXISTS*/ %s;\n",status->current_schema_quoted));
    }
    else
    {
      row= mysql_fetch_row(result);
      if (!row)
      {
	g_free(stmt);
	*error= MYX_BACKUP_MYSQL_ERROR;
	return -1;
      }
      database_name= myx_convert_dbstr_utf8(status->mysql, row[1]);

      SAFE_IO(fprintf(status->sql_file, "%s;\n", database_name));

      g_free(database_name);
    }
    g_free(stmt);


    SAFE_IO(fprintf(status->sql_file, "USE %s;\n", status->current_schema_quoted));
  }

  return 0;
}

static int write_sql_file_header(MYX_BS_STATUS *status, MYX_BACKUP_ERROR *error)
{
  if (status->options & MYX_B_COMMENT)
  {
    SAFE_IO(fprintf(status->sql_file, "-- MySQLAdministrator dump %s\n--\n", ADMIN_DUMP_VERSION));
    //    fprintf(status->sql_file, "-- Host: %s    Database: %s\n", current_host ? current_host : "localhost", db_name ? db_name : "");
    SAFE_IO(fputs("-- ------------------------------------------------------\n", status->sql_file));
    SAFE_IO(fprintf(status->sql_file, "-- Server version\t%s\n", mysql_get_server_info(status->mysql)));
    SAFE_IO(fprintf(status->sql_file, "\n\n\n"));
  }

  //fprintf(status->sql_file,"\n/*!40101 SET @OLD_CHARACTER_SET_CLIENT=@@CHARACTER_SET_CLIENT, CHARACTER_SET_CLIENT=%s */;\n", DEFAULT_CHARSET);
  SAFE_IO(fprintf(status->sql_file,"/*!40014 SET @OLD_UNIQUE_CHECKS=@@UNIQUE_CHECKS, UNIQUE_CHECKS=0 */;\n"));
  SAFE_IO(fprintf(status->sql_file,"/*!40014 SET @OLD_FOREIGN_KEY_CHECKS=@@FOREIGN_KEY_CHECKS, FOREIGN_KEY_CHECKS=0 */;\n"));
  if (status->options & MYX_B_ANSI_QUOTES)
    SAFE_IO(fprintf(status->sql_file,"/*!40101 SET @OLD_SQL_MODE=@@SQL_MODE, SQL_MODE='ANSI_QUOTES,NO_AUTO_VALUE_ON_ZERO' */;\n"))
  else  
    SAFE_IO(fprintf(status->sql_file,"/*!40101 SET @OLD_SQL_MODE=@@SQL_MODE, SQL_MODE=NO_AUTO_VALUE_ON_ZERO */;\n"))
  
  return 0;
}

static int write_sql_file_footer(MYX_BS_STATUS *status, MYX_BACKUP_ERROR *error)
{
  SAFE_IO(fprintf(status->sql_file,"/*!40101 SET SQL_MODE=@OLD_SQL_MODE */;\n"));
  SAFE_IO(fprintf(status->sql_file,"/*!40014 SET FOREIGN_KEY_CHECKS=@OLD_FOREIGN_KEY_CHECKS */;\n"));
  SAFE_IO(fprintf(status->sql_file,"/*!40014 SET UNIQUE_CHECKS=@OLD_UNIQUE_CHECKS */;\n"));
  //fprintf(status->sql_file,"/*!40101 SET CHARACTER_SET_CLIENT=@OLD_CHARACTER_SET_CLIENT */;\n");

  return 0;
}

static int write_create_statement_to_file(MYX_BS_STATUS *status, MYX_BACKUP_ERROR *error)
{
  char *stmt, *create_statement;
  MYSQL_ROW row;
  MYSQL_RES *result;

  /* make sure show create table's output is quoted */
  if (myx_mysql_query(status->mysql, "SET SQL_QUOTE_SHOW_CREATE=1"))
  {
    *error= MYX_BACKUP_MYSQL_ERROR;
    return -1;
  }

  stmt= g_strconcat("SHOW CREATE TABLE ", status->current_schema_quoted, ".", status->current_table_quoted, NULL);
  if ((myx_mysql_query(status->mysql, stmt)) || !(result= mysql_store_result(status->mysql)) )
  {
    g_free(stmt);
    *error= MYX_BACKUP_MYSQL_ERROR;
    return -1;
  }
  row= mysql_fetch_row(result);
  if (!row)
  {
    g_free(stmt);
    *error= MYX_BACKUP_MYSQL_ERROR;
    return -1;
  }
  g_free(stmt);

  if (status->options & MYX_B_ADD_DROP_TABLE)
    SAFE_IO(fprintf(status->sql_file, "\nDROP TABLE IF EXISTS %s;\n", status->current_table_quoted));

  /* TODO: check if the quotes are set right */

  create_statement= myx_convert_dbstr_utf8(status->mysql, row[1]); /* row[1] has the create-statement */

  SAFE_IO(fprintf(status->sql_file,"%s;\n", create_statement)); 

  g_free(create_statement);

  mysql_free_result(result);

  return 0;
}


static int write_row_to_file(MYX_BS_STATUS *status, MYSQL_ROW row, MYX_BACKUP_ERROR *error)
{
  unsigned int num_fields;
  unsigned long *lengths;
  unsigned int i;
  unsigned char *insert_beg, *value_stmt;
//  unsigned char *tmp;
  MYSQL_FIELD *fields;

  value_stmt= g_strdup("");

  num_fields = mysql_num_fields(status->mysql_result);
  lengths = mysql_fetch_lengths(status->mysql_result);
  fields = mysql_fetch_fields(status->mysql_result);

  if (status->options & MYX_B_COMPLETE_INSERTS)
  {
    insert_beg= g_strconcat("INSERT INTO ", status->current_table_quoted, " (", NULL); //" VALUES ", NULL);
    
    for(i = 0; i < num_fields; i++)
    {
      char *quoted= g_malloc0((unsigned int)strlen(fields[i].name)*2+3);
      quote_name(fields[i].name, quoted, status->quote_char);

      insert_beg= myx_gstr_append(insert_beg, g_strconcat(quoted,",",NULL));

      g_free(quoted);
    }
    insert_beg[strlen(insert_beg)-1]= ')';

    insert_beg= myx_gstr_append(insert_beg, " VALUES ");
  }
  else
  {
    insert_beg= g_strconcat("INSERT INTO ", status->current_table_quoted," VALUES ",NULL);
  }

  if (! status->extended_insert)
  {
    status->extended_insert= g_strdup(insert_beg);
  }

  for(i = 0; i < num_fields; i++)
  {
    bigint length= lengths[i];
    MYSQL_FIELD *field= fields+i;

    if (i == 0)
    {
      value_stmt= myx_gstr_append(value_stmt, "(");
    }
    else
    {
      value_stmt= myx_gstr_append(value_stmt, ",");
    }

    if (row[i]) /* if != NULL */
    {
      if (length) /* if != empty string */
      {
        if (!IS_NUM_FIELD(field))
        {
          char *tmp_str= g_malloc((unsigned int)length*2+1+2);
          char *row_value= myx_convert_dbstr_utf8(status->mysql, row[i]);
          unsigned int len= mysql_real_escape_string(status->mysql,tmp_str+1, row_value, (unsigned int)length);
          
          g_free(row_value);

	  tmp_str[0]= '\'';
	  tmp_str[len+1]= '\'';
	  tmp_str[len+2]= '\0';
	  value_stmt= myx_gstr_append(value_stmt, tmp_str);

          g_free(tmp_str);
        }
        else
        {
	  if ( !strcmp(row[i],"inf") || !strcmp(row[i],"-inf") || !strcmp(row[i],"nan") )
	  {
	    value_stmt= myx_gstr_append(value_stmt, "NULL");
	  }
	  else
	  {
	    if (field->type == FIELD_TYPE_DECIMAL)
	    {
	      value_stmt= myx_gstr_append(value_stmt,g_strconcat("'",row[i],"'",NULL));
	    }
	    else
	      value_stmt= myx_gstr_append(value_stmt, row[i]);
	  }
        }
      }
      else
      {
        //write ''
	value_stmt= myx_gstr_append(value_stmt, "''");
      }
    }
    else
    {
      //write NULL
      value_stmt= myx_gstr_append(value_stmt, "NULL");
    }
  }
  /* end for */

  value_stmt= myx_gstr_append(value_stmt, ")");

  if (! (status->options & MYX_B_NO_EXTENDED_INSERT) ) /*extended inserts are not always written out*/
  { 
    if (strlen(status->extended_insert) < 1000)
    {
      //tmp= status->extended_insert;
      //status->extended_insert= g_strconcat(status->extended_insert, " ", value_stmt, ",", NULL);
      //g_free(tmp);
      status->extended_insert= myx_gstr_append(status->extended_insert, value_stmt);
      status->extended_insert= myx_gstr_append(status->extended_insert, ",");
    }
    else
    {
      /* we write it out */
/*      tmp= status->extended_insert;
	status->extended_insert= g_strconcat(status->extended_insert, value_stmt, ";", NULL);
      g_free(tmp);
*/
      status->extended_insert= myx_gstr_append(status->extended_insert, value_stmt);
      status->extended_insert= myx_gstr_append(status->extended_insert, ";");

      SAFE_IO(fprintf(status->sql_file, "%s\n",status->extended_insert));
      g_free(status->extended_insert);
      status->extended_insert= NULL;
    }
  }
  else
  {
    SAFE_IO(fprintf(status->sql_file, "%s %s;\n",insert_beg, value_stmt));
  }

  g_free(insert_beg);
  g_free(value_stmt);

  return 0;
}

/* from mysqldump.c but slightly modified */
//mysql's native quote char is '`' (also works when mysql is running
//in ansi mode)
//the ansi quote char is '"'
static char *quote_name(const char *name, char *buff, char quote_char)
{
  char *to= buff;

  *to++= quote_char;
  while (*name)
  {
    if (*name == quote_char)
      *to++= quote_char;
    *to++= *name++;
  }
  to[0]= quote_char;
  to[1]= 0;
  return buff;
}
/* quote_name */

/* we assume that we are only called with correctly
 * quoted strings or strings that are not quoted
 * TODO: change that.
 */
static char* unquote_name(char *table_name)
{
  int write_back= -1;
  char quote_char;
  char *beginning;

  beginning= table_name;

  if (*table_name == '"' || *table_name == '`')
    quote_char= *table_name++;
  else
    return table_name; /* the string is not quoted */

  for(;*table_name; table_name++)
  {
    if (*table_name == quote_char && table_name[-1] == quote_char)
      write_back--;

    table_name[write_back]= table_name[0];
  }

  table_name[write_back-1]= '\0';

  return beginning;
}

static int comfort_query(MYX_RBS_STATUS *rs, char *buffer, int buffer_len, MYX_BACKUP_ERROR *error)
{
  char *warning;

  if (buffer[buffer_len-1] == ';')
  {
    buffer_len -= 1;
  }
  if (myx_mysql_real_query(rs->mysql, buffer,buffer_len) )
  {
    if (! (rs->options &  MYX_RBS_FORCE))
    {
      //fprintf(stderr, "Error: Error while executing this query: %s\nThe server has returned this error message: %s\n", buffer, mysql_error(rs->mysql));
      *error= MYX_BACKUP_MYSQL_ERROR;
      return -1;
    }

    warning= g_strconcat("Warning: Error while executing this query:", buffer ,"\nThe server has returned this error message:", mysql_error(rs->mysql), NULL);
    if (rs->report_warning)
      (*rs->report_warning)(warning, rs->report_warning_data);
    else
      fprintf(stderr, "%s\n",warning);
    g_free(warning);
  }

  return 0;
}

/* sends the query to the database, issueing a use database if necessary before */
static int interact_with_server(MYX_RBS_STATUS *rs, char *buffer, int buffer_len,  MYX_BACKUP_ERROR *error)
{
  if ((!rs->target_schema)) /* take the target-information from the sql-file */
  {
    /* do nothing
    char *tmp;

    tmp= g_strconcat("USE ", rs->current_schema_in_sqlfile, NULL);

    if (mysql_query(rs->mysql,tmp)) // Error executing USE
    {
      char *warning;
      
      if (! (rs->options & MYX_RBS_FORCE) )
      {
        g_free(tmp);
        *error= MYX_BACKUP_MYSQL_ERROR;
        //fprintf(stderr, "Error: Could not change the db.\n The server has returned this error message: %s\n", mysql_error(rs->mysql));
        return -1;
      }
      warning= g_strconcat("Warning: Could not change the db.\n The server has returned this error message:", mysql_error(rs->mysql), NULL);
      if (rs->report_warning)
        (*rs->report_warning)(warning, rs->report_warning_data);
      else
        fprintf(stderr,"%s", warning);
      g_free(warning);
      
    }
    else
    {
      g_free(rs->current_schema_in_db);
      rs->current_schema_in_db= g_strdup(rs->current_schema_in_sqlfile);
    } */
  }
  else if ( strcmp(rs->current_schema_in_db, rs->target_schema) ) /* use the supplied target-information */
  {
    char *tmp;

    tmp= g_strconcat("USE ", rs->target_schema, NULL);

    if (myx_mysql_query(rs->mysql,tmp)) /* Error executing USE */
    {
      if ( (mysql_errno(rs->mysql) == 1049) && !(rs->options & MYX_RBS_DONT_CREATE_TARGETS)) /* Unknown database */
      {
        char *tmp2= g_strconcat("CREATE DATABASE ", rs->target_schema, NULL);

        if ( comfort_query(rs, tmp2, (int)strlen(tmp2), error ) )
        {
          g_free(tmp);
          g_free(tmp2);
          return -1;
        }
        if ( comfort_query(rs, tmp, (int)strlen(tmp), error ) )
        {
          g_free(tmp);
          g_free(tmp2);
          return -1;
        }
        g_free(tmp2);
      }
      else
      {
        char *warning;

        if (! (rs->options & MYX_RBS_FORCE) )
        {
          g_free(tmp);
          *error= MYX_BACKUP_MYSQL_ERROR;
          //fprintf(stderr, "Error: Could not change the db.\n The server has returned this error message: %s\n", mysql_error(rs->mysql));
          return -1;
        }
        warning= g_strconcat("Warning: Could not change the db.\n The server has returned this error message:", mysql_error(rs->mysql), NULL);
        if (rs->report_warning)
          (*rs->report_warning)(warning, rs->report_warning_data);
        else
          fprintf(stderr,"%s", warning);
        g_free(warning);
      }
    }

    g_free(tmp);

    g_free(rs->current_schema_in_db);
    rs->current_schema_in_db= g_strdup(rs->target_schema);
  }
  /* end if strcmp */

  //go ahead and send a query with this line
  return ( comfort_query(rs, buffer,buffer_len, error) );
}

static void free_backup_table_content(MYX_BACKUP_TABLE *t)
{
  g_free(t->schema);
  g_free(t->catalog);
  g_free(t->table);
}

/* returns >0 if the given table is part of the backup-content */
static int find_table(MYX_BACKUP_CONTENT *b, const char *table_name, const char *schema_name, const char *catalog_name)
{
  unsigned int i;

  for (i=0; i < b->tables_num; i++)
  {
    //Ignore catalogs for now
    if (! (strcmp(b->tables[i].table, table_name)) && !(strcmp(b->tables[i].schema, schema_name))/* && !(strcmp(b->tables[i].catalog, catalog_name))*/ )
      return 1;
  }

  return 0;
}


/* Rexexp functions */

#define O_VECTOR_COUNT 6

static char* check_statement(const char *line, int line_len, pcre *re, pcre_extra *pe)
{
  int o_vector[O_VECTOR_COUNT];
  const char *ret_val;
  int rc;
  char *return_value;

  if ( (rc= pcre_exec(re,pe, line, line_len,0,0,o_vector, O_VECTOR_COUNT) ) > 0)
  {
    pcre_get_substring(line, o_vector, rc, 1, &ret_val);
    return_value= g_strdup(ret_val);
    pcre_free((char*)ret_val);
    return return_value;
  }

  return NULL;
}

/*returns the table-name or NULL if this is no create-statement*/
static int prepare_line_is_create_statement(pcre **re, pcre_extra **pe)
{
  const char *error;
  int erroffset;

  if (! (*re = pcre_compile("create\\s+table\\s+(\\S+)", /* the pattern */
                            PCRE_CASELESS|PCRE_ANCHORED, /* default options */
                            &error, /* for error message */
                            &erroffset, /* for error offset */
                            NULL)) ) /* use default character tables */
  {
    fprintf(stderr, "%s\n",error);
    return 1;
  }

  if (pe)
  {
    *pe = pcre_study(
                     *re, /* result of pcre_compile() */
                     0, /* no options exist */
                     &error);
  }
  return 0;
}

static void finalize_line_is_create_statement(pcre *re, pcre_extra *pe)
{
  if (re) pcre_free(re);
  if (pe) pcre_free(pe);
}

static int prepare_line_is_db_use_statement(pcre **re, pcre_extra **pe)
{
  const char *error;
  int erroffset;

  if (! (*re = pcre_compile("use\\s+(\\S+)\\s*;", /* the pattern */
                            PCRE_CASELESS|PCRE_ANCHORED, /* default options */
                            &error, /* for error message */
                            &erroffset, /* for error offset */
                            NULL)) ) /* use default character tables */
  {
    fprintf(stderr, "%s\n",error);
    return 1;
  }

  if (pe)
  {
    *pe = pcre_study(
                     *re, /* result of pcre_compile() */
                     0, /* no options exist */
                     &error);
  }
  return 0;
}

static void finalize_line_is_db_use_statement(pcre *re, pcre_extra *pe)
{
  if (re) pcre_free(re);
  if (pe) pcre_free(pe);
}

static WORKING_ENVIRONMENT* new_working_environment(void)
{
  WORKING_ENVIRONMENT *we;

  we= g_malloc0(sizeof(WORKING_ENVIRONMENT));
  we->state= 1;
  we->bytes_read= 0;

  return we;
}

static void free_working_environment(WORKING_ENVIRONMENT *we)
{
  g_free(we);
}


/*
void test_get_next_sql_statement(void)
{
  WORKING_ENVIRONMENT *we;
  MYX_BACKUP_ERROR error_code;
  char buffer[255];
  int buffer_len= 255;
  FILE *file;
  char *bp= buffer;
  GDir* dir;
  const char *filename;

  dir= g_dir_open("sql-files",0,NULL);

  while (filename= g_dir_read_name (dir))
  {
    FILE *stmt_file;
    
    filename= g_strconcat("sql-files/",filename, NULL);

    we= new_working_environment();

    if (! (file=fopen(filename, "r")) )
    {
      fprintf(stderr, "Error opening file.\n");
      exit(1);
    }

    if (! (stmt_file=fopen(g_strconcat(filename,"_statements",NULL), "w")) )
    {
      fprintf(stderr, "Error opening statements file.\n");
      exit(1);
    } 

    while (get_next_sql_statement(we, file, &bp,&buffer_len,0, &error_code))
    {
      fprintf(stmt_file,"Statement: %s\n",buffer);
    }
    fclose(file);
    fclose(stmt_file);
    free_working_environment(we);
    g_free(filename);
  }

  g_dir_close(dir);
} */


//N = newline, E= any other character
static int state_machine[15][12] =
{
  //      ; ' " # - / * N \ E ! `
/*0*/    {1,3,2,6,4,7,0,0,0,0,0,14},
/*1*/    {13,13,13,13,13,13,13,13,13,13,13,13},
/*2*/    {2,2,0,2,2,2,2,2,11,2,2,2},
/*3*/    {3,0,3,3,3,3,3,3,10,3,3,3},
/*4*/    {1,3,2,6,5,7,0,0,0,0,0,14},
/*5*/    {5,5,5,5,5,5,5,0,5,5,5,5},
/*6*/    {6,6,6,6,6,6,6,0,6,6,6,6},
/*7*/    {1,3,2,6,4,7,8,0,0,0,0,14},
/*8*/    {9,9,9,9,9,9,9,9,9,9,0,9},
/*9*/    {9,9,9,9,9,9,12,9,9,9,9,9},
/*10*/   {3,3,3,3,3,3,3,3,3,3,3,3},
/*11*/   {2,2,2,2,2,2,2,2,2,2,2,2},
/*12*/   {9,9,9,9,9,0,12,9,9,9,9,9},
/*13*/   {13,13,13,13,13,13,13,13,13,13,13,13},
/*14*/   {14,14,14,14,14,14,14,14,14,14,0}
};

/*
 *----------------------------------------------------------------------
 * Returns the next sql-statement
 *
 * SYNOPSIS
 *
 * DESCRIPTION
 *
 * RETURN VALUE
 *
 * NOTES
 *----------------------------------------------------------------------
 */
/* TODO: check if this is multibyte/utf-8 safe -alfredo */
static int get_next_sql_statement(WORKING_ENVIRONMENT *we, FILE *sql_file, char **buffer, int *buffer_len, int resize_buffer, MYX_BACKUP_ERROR *error)
{
  int h;
  int c;
  int state;
  
  state= we->state;
  
  //fast-forward
  while ( (state!=1) && ((c= fgetc(sql_file))!=EOF) )
  {
    we->bytes_read++;
    state= make_state_transition(state, c);
  }
  
  /* read in (the beginning of) the next sql-statement */
  if (state == 1)
    state= 0;

  
  h=0;
  while ( (state!=1) && (resize_buffer || h< *buffer_len-1) && ((c= fgetc(sql_file))!=EOF) )
  {
    we->bytes_read++;

    if ((resize_buffer) && (h == *buffer_len - 2))
    {
      gchar *tmp;

      /* increase buffer resize */
      tmp= g_try_realloc(*buffer, *buffer_len + PACKET_RESIZE_INCREMENT);
      if (!tmp)
      {
        *error= MYX_BACKUP_MALLOC_FAILED;
        return 0;
      }
      *buffer= tmp;
      *buffer_len += PACKET_RESIZE_INCREMENT;
    }

    /* Record the character if we are in no comment section */
    if (state != 6 && state != 5 && state != 9 && state != 12)
    {
      //write
      if ( !(h==0 && isspace(c)) ) /* ignore beginning white-space */
        (*buffer)[h++]=c; 
    }

    /* If this character triggers a comment, delete the comment-beginning */
    if (state == 4 && c == '-')
    {
      h -= 2;      
    }
    else if (state == 0 && c == '#')
    {
      h -= 1;
    }
    else if (state == 8 && c != '!')
    {
      h -= 3;
    }

    
    state= make_state_transition(state, c);
  
  }

  (*buffer)[h]= '\0';

  we->state= state;

  return h;
}

static int make_state_transition(int state, int c)
{
  
  switch(c){

    case ';':
      state=state_machine[state][0];
      break;
    case '\'':
      state=state_machine[state][1];
      break;
    case '"':
      state=state_machine[state][2];
      break;
    case '#':
      state=state_machine[state][3];
      break;
    case '-':
      state=state_machine[state][4];
      break;
    case '/':
      state=state_machine[state][5];
      break;
    case '*':
      state=state_machine[state][6];
      break;
    case '\n':
      state=state_machine[state][7];
      break;
    case '\\':
      state=state_machine[state][8];
      break;
    case '!':
      state=state_machine[state][10];
      break;
    case '`':
      state=state_machine[state][11];
      break;
    default:
      state=state_machine[state][9];
      break;
    } 
  
  return state;
}
