/*
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "myx_admin_library.h"
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <libxml/tree.h>

//#include <assert.h>
#include <stdlib.h>
#include <myx_xml_aux_functions.h>


/*
 * Forward declarations
 */
static MYX_VARIABLES_LISTING* read_in_variables_listing(const xmlNodePtr node);
static void read_in_group(const xmlNodePtr node, MYX_VARIABLES_GROUP *group);
static void read_in_subgroup(const xmlNodePtr node, MYX_VARIABLES_SUBGROUP *subgroup);
static void read_in_variable_element(const xmlNodePtr node, MYX_VARIABLE_ELEMENT *element);

static void free_group_content(MYX_VARIABLES_GROUP *group);
static void free_subgroup_content(MYX_VARIABLES_SUBGROUP *subgroup);
static void free_variable_element_content(MYX_VARIABLE_ELEMENT *variable);

static void sort_groups_and_subgroups(MYX_VARIABLES_LISTING *variables_listing);

/*
 * Public functions
 */
/*int main()
{
  MYX_VARIABLES_LISTING* l;
  MYX_ADMIN_LIB_ERROR error_code;

  l = myx_get_variables_listing("../xmlfiles/status_variables.xml", &error_code);
  if (l == NULL)
  {
    return error_code;
  }
  else
  {
    myx_free_variables_listing(l);
    return 0;
  }
}*/


MYX_VARIABLES_LISTING* myx_get_variables_listing(unsigned char* filename, MYX_ADMIN_LIB_ERROR *error_code)
{
  xmlDocPtr doc;
  xmlParserCtxtPtr ctxt;
  xmlNodePtr root;
  //xmlNodePtr cur;
  //int i;
  MYX_VARIABLES_LISTING *variables_listing;

  *error_code = MYX_ADMIN_NO_ERROR;

  if (! file_exists(filename) )
  {
    *error_code = MYX_ADMIN_ERROR_CANT_OPEN_FILE;
    return NULL;
  }

  /* create a parser context */
  ctxt = xmlNewParserCtxt();
  if (ctxt == NULL) 
  {
    return NULL;
  }
  /* parse the file, activating the DTD validation option */
  doc = xmlCtxtReadFile(ctxt, filename, NULL, 0);

  /* check if parsing suceeded */
  if (doc == NULL) 
  {
    *error_code = MYX_ADMIN_XML_PARSE_ERROR;
    xmlFreeParserCtxt(ctxt);
    return NULL;
  } 
  else 
  {
    /* check if validation suceeded */
    if (ctxt->valid == 0)
    {
      *error_code = MYX_ADMIN_XML_NO_VALID_DOCUMENT;
      xmlFreeParserCtxt(ctxt);	    
      xmlFreeDoc(doc);
      return NULL;
    }
  }
  /* free up the parser context */
  xmlFreeParserCtxt(ctxt);

      
  root = xmlDocGetRootElement(doc);
    
  if (root == NULL) {
    *error_code = MYX_ADMIN_XML_EMPTY_DOCUMENT;
    xmlFreeDoc(doc);
    return NULL;
  }
    
  variables_listing = read_in_variables_listing(root);

  sort_groups_and_subgroups(variables_listing);

  xmlFreeDoc(doc);
    
  return variables_listing;    
}


int myx_free_variables_listing(MYX_VARIABLES_LISTING* variables_listing)
{
  unsigned int i;

  if (variables_listing)
  {
    for (i=0; i < variables_listing->groups_num; i++)
    {
      free_group_content(variables_listing->groups +i);
    }
    free(variables_listing->groups);
    free(variables_listing);
  }

  return 0;
}



/*
 * Private functions
 */ 



/*read_in helper functions */

static MYX_VARIABLES_LISTING* read_in_variables_listing(const xmlNodePtr node)
{
  MYX_VARIABLES_LISTING *variables_listing;
  int i;
  xmlNodePtr cur;

  variables_listing = calloc(1, sizeof(MYX_VARIABLES_LISTING) );
 
  variables_listing->groups_num = get_child_count(node,"group");
  if (variables_listing->groups_num  > 0)
  {
    variables_listing->groups= calloc(variables_listing->groups_num,sizeof(MYX_VARIABLES_GROUP));

    i=0;
    cur = node->children;
    while (cur != NULL) 
    {
      /* read a single group*/
      if (!xmlStrcmp(cur->name, "group"))
      {
	read_in_group(cur, variables_listing->groups+i);
	i++;
      }

      cur = cur->next;
    }
  }

  return variables_listing;
}


static void read_in_group(const xmlNodePtr node, MYX_VARIABLES_GROUP *group)
{
  int i;
  xmlNodePtr cur;

  /* read in attributes */
  group->pos= atoi_and_free(xmlGetProp(node, "pos"));
  group->name = xmlGetProp(node, "name");
  group->caption_id = xmlGetProp(node, "caption_id");

  /* read variables */
  group->variables_num = get_child_count(node, "variable");
  if (group->variables_num > 0)
  { 
    group->variables = calloc(group->variables_num, sizeof(MYX_VARIABLE_ELEMENT));

    i= 0;
    cur= node->children;
    while (cur)
    {
      if (!xmlStrcmp(cur->name, "variable"))
      {
	read_in_variable_element(cur, group->variables+i);
	i++;
      }

      cur=cur->next;
    }
  }
  /* read in subgroups */
  group->subgroups_num = get_child_count(node, "subgroup");
  if (group->subgroups_num > 0) 
  {
    group->subgroups= calloc(group->subgroups_num, sizeof(MYX_VARIABLES_SUBGROUP));

    i= 0;
    cur= node->children;
    while (cur)
    {
      if (!xmlStrcmp(cur->name, "subgroup"))
      {
	read_in_subgroup(cur, group->subgroups+i);
	i++;
      }

      cur=cur->next;
    }
  }
}


static void read_in_subgroup(const xmlNodePtr node, MYX_VARIABLES_SUBGROUP *subgroup)
{
  int i;
  xmlNodePtr cur;

  /* read in attributes */
  subgroup->pos= atoi_and_free(xmlGetProp(node, "pos"));
  subgroup->name = xmlGetProp(node, "name");
  subgroup->caption_id = xmlGetProp(node, "caption_id");

  /* read variables */
  subgroup->variables_num = get_child_count(node, "variable");
  if (subgroup->variables_num > 0) 
  {
    subgroup->variables = calloc(subgroup->variables_num, sizeof(MYX_VARIABLE_ELEMENT));

    i= 0;
    cur= node->children;
    while (cur)
    {
      if (!xmlStrcmp(cur->name, "variable"))
      {
	read_in_variable_element(cur, subgroup->variables+i);
	i++;
      }

      cur=cur->next;
    } 
  }
}

static void read_in_variable_element(const xmlNodePtr node, MYX_VARIABLE_ELEMENT *element)
{
  xmlChar *tmp;

  element->mysql_id = xmlGetProp(node, "mysql_id");
  element->desc_id = xmlGetProp(node, "desc_id");
  tmp= xmlGetProp(node,"editable");
  if (tmp && !xmlStrcmp(tmp, "yes"))
    element->editable= 1;
  else
    element->editable= 0;
  if (tmp) xmlFree(tmp);
}


/* free helper functions */

static void free_group_content(MYX_VARIABLES_GROUP *group)
{
  unsigned int i;

  xmlFree(group->name);
  xmlFree(group->caption_id);

  for(i=0; i < group->variables_num; i++)
  {
    free_variable_element_content(group->variables+i);
  }
  free(group->variables);

  for (i=0; i < group->subgroups_num; i++)
  {
    free_subgroup_content(group->subgroups+i);
  }
  free(group->subgroups);
}

static void free_subgroup_content(MYX_VARIABLES_SUBGROUP *subgroup)
{
  unsigned int i;

  xmlFree(subgroup->name);
  xmlFree(subgroup->caption_id);

  for(i=0; i < subgroup->variables_num; i++)
  {
    free_variable_element_content(subgroup->variables+i);
  }
  free(subgroup->variables);
}

static void free_variable_element_content(MYX_VARIABLE_ELEMENT *variable)
{
  xmlFree(variable->mysql_id);
  xmlFree(variable->desc_id);
}


/*Functions for sorting the groups*/
static int compare_groups(const void *a, const void *b)
{
  MYX_VARIABLES_GROUP *g1 = (MYX_VARIABLES_GROUP *) a;
  MYX_VARIABLES_GROUP *g2 = (MYX_VARIABLES_GROUP *) b;

  if (g1->pos < g2->pos)
  {
    return -1;
  }
  else if (g1->pos > g2->pos)
  {
    return 1;
  }
  else
  {
    return 0;
  }
}


static int compare_subgroups(const void *a, const void *b)
{
  MYX_VARIABLES_SUBGROUP *g1 = (MYX_VARIABLES_SUBGROUP *) a;
  MYX_VARIABLES_SUBGROUP *g2 = (MYX_VARIABLES_SUBGROUP *) b;

  if (g1->pos < g2->pos)
  {
    return -1;
  }
  else if (g1->pos > g2->pos)
  {
    return 1;
  }
  else
  {
    return 0;
  }
}

static void sort_groups_and_subgroups(MYX_VARIABLES_LISTING *variables_listing)
{
  unsigned int i;

  if (variables_listing->groups)
  {
    /*sort the groups*/
    qsort(variables_listing->groups, variables_listing->groups_num, sizeof(MYX_VARIABLES_GROUP), compare_groups);

    /*sort the subgroups contained in each group*/
    for (i=0; i < variables_listing->groups_num; i++)
    {
      if (variables_listing->groups[i].subgroups)
	qsort(variables_listing->groups[i].subgroups, variables_listing->groups[i].subgroups_num, sizeof(MYX_VARIABLES_SUBGROUP), compare_subgroups);
    }
  }
}
