unit ApplicationDataModule;

interface

uses
  SysUtils, Classes, ComCtrls, ImgList, Controls, Forms, Contnrs, Windows,
  ExtCtrls, AuxFuncs, PNGImage, StdCtrls,
  myx_public_interface, Options, MyxError, StrUtils;

type
  TMYXAdminOptions = Class;

  TApplicationDM = class(TDataModule)
    ServiceStatusImageList: TImageList;
    AdminTree16ImageList: TImageList;
    AdminTree24ImageList: TImageList;
    Admin48ImageList: TImageList;
    CatalogImageList: TImageList;
    SectionImageList: TImageList;
    procedure DataModuleCreate(Sender: TObject);
    procedure DataModuleDestroy(Sender: TObject);

    procedure OnApplicationException(Sender: TObject; E: Exception);
    function CheckFiles: Boolean;
    procedure LoadTranslations;
    procedure LoadOptions;
    function GetLastFileDialogPaths(DlgName: string): string;
    procedure SetLastFileDialogPaths(DlgName: string; Path: string);
  private
    { Private declarations }
  public
    { Public declarations }
    ApplicationIsTerminating: Boolean;

    pmyx_trans: PMYX_TRANS;

    Options: TMYXAdminOptions;
  end;

  TMYXAdminOptions = Class(TMyxOptions)
    constructor Create; reintroduce; overload;
    destructor Destroy; override;

    procedure LoadOptions; override;
    procedure StoreOptions; override;
  public
    //Global
    ShowTipOfDay,
    StoreWindowsPositions: Boolean;

    ShowOnlyServiceSections: Boolean;
    MySQLInstallPath: string;
    MySQLVersion: double;

    ShowUserGlobalPrivileges: Boolean;
    ShowUserTableColumnPrivileges: Boolean;

    StartSection: integer;
    SectionSidebarWidth: integer;
    SectionSidebarHidden: Boolean;

    UsePeakLevel: Boolean;
    ResetPeakLevel: Boolean;
    PeakLevelResetTicks: integer;

    LastFileDialogPaths: TStringList;
  end;

  EMyxAdminLibError = class(EMyxLibraryError)
    constructor Create(ErrorNr: integer; ErrorTxt: string); override;
  end;

var
  ApplicationDM: TApplicationDM;

implementation

uses Main, ConnectToInstance;

{$R *.dfm}

procedure TApplicationDM.DataModuleCreate(Sender: TObject);
begin
  //DLL Version check

  //Exception Handling
  Application.OnException:=OnApplicationException;

  ApplicationIsTerminating:=False;
end;

procedure TApplicationDM.DataModuleDestroy(Sender: TObject);
begin
  if(pmyx_trans<>nil)then
    myx_free_trans(pmyx_trans);

  Options.Free;
end;

procedure TApplicationDM.OnApplicationException(Sender: TObject; E: Exception);
begin
  if(not(Application.Terminated))then
  begin
    if(E is EMyxAdminLibError)then
    begin
      ShowModalDialog(Application.Title+' Admin Library Error', E.Message+#13#10#13#10+
        'Admin Library Error Nr.'+EMyxAdminLibError(E).ErrorNr+#13#10+
        EMyxAdminLibError(E).Message, myx_mtError, 'OK');
    end
    else
      ShowError(E);
  end;
end;

constructor EMyxAdminLibError.Create(ErrorNr: integer; ErrorTxt: string);
begin
  //self.Message:=Message;
  self.ErrorNr:=IntToStr(ErrorNr);

  case ErrorNr of
    1:
      self.Message:=Format('The File %s cannot be opened.', [ErrorTxt]);
    2:
      self.Message:=Format('An error occured while parsing the XML file %s.', [ErrorTxt]);
    3:
      self.Message:=Format('An error occured while validating the XML file %s.', [ErrorTxt]);
    4:
      self.Message:=Format('The XML file %s is empty.', [ErrorTxt]);
    5:
      self.Message:=Format('An error occured while parsing the INI file %s.', [ErrorTxt]);
  else
    self.Message:='';
  end;
end;

function TApplicationDM.CheckFiles: Boolean;
var MissingFiles: string;
begin
  CheckFiles:=True;

  MissingFiles:='';

  if(Not(FileExists(MYXCommonOptions.XMLDir+'mysqlx_translations_general.xml')))then
    MissingFiles:=MissingFiles+
      MYXCommonOptions.XMLDir+'mysqlx_translations_general.xml'+#13#10;
  if(Not(FileExists(MYXCommonOptions.XMLDir+'mysqlx_translations_administrator.xml')))then
    MissingFiles:=MissingFiles+
      MYXCommonOptions.XMLDir+'mysqlx_translations_administrator.xml'+#13#10;
  if(Not(FileExists(MYXCommonOptions.XMLDir+'mysqlx_startup_variables_description.xml')))then
    MissingFiles:=MissingFiles+
      MYXCommonOptions.XMLDir+'mysqlx_startup_variables_description.xml'+#13#10;
  if(Not(FileExists(MYXCommonOptions.XMLDir+'mysqlx_status_variables.xml')))then
    MissingFiles:=MissingFiles+
      MYXCommonOptions.XMLDir+'mysqlx_status_variables.xml'+#13#10;
  if(Not(FileExists(MYXCommonOptions.XMLDir+'mysqlx_system_variables.xml')))then
    MissingFiles:=MissingFiles+
      MYXCommonOptions.XMLDir+'mysqlx_system_variables.xml'+#13#10;
  if(Not(FileExists(MYXCommonOptions.XMLDir+'mysqlx_startup_variables_description.dtd')))then
    MissingFiles:=MissingFiles+
      MYXCommonOptions.XMLDir+'mysqlx_startup_variables_description.dtd'+#13#10;

  if(MissingFiles<>'')then
  begin
    ShowModalDialog('Files missing.',
      'The file following files cannot be found. '+#13#10+
      'Please check the file path and the existence of the file.'+#13#10#13#10+
      MissingFiles);

    CheckFiles:=False;
  end;
end;

procedure TApplicationDM.LoadTranslations;
var error: MYX_LIB_ERROR;
begin
  try
    pmyx_trans:=myx_init_trans(
      PChar(AnsiToUtf8(MYXCommonOptions.XMLDir+'mysqlx_translations_general.xml')),
      PChar(AnsiToUtf8(MYXCommonOptions.XMLDir+'mysqlx_translations_administrator.xml')),
      PChar(AnsiToUtf8(MYXCommonOptions.Language)), @error);
    if(pmyx_trans=nil)or(error<>MYX_NO_ERROR)then
      raise EMyxError.Create('Could not load translations from files '#13#10+
        MYXCommonOptions.XMLDir+'mysqlx_translations_general.xml and '#13#10+
        MYXCommonOptions.XMLDir+'mysqlx_translations_administrator.xml.'#13#10+
        'Error Nr.: '+IntToStr(Ord(error)));
  except
    pmyx_trans:=nil;
    raise;
  end;
end;

procedure TApplicationDM.LoadOptions;
begin
  //Load Options
  Options:=TMYXAdminOptions.Create;
end;

function TApplicationDM.GetLastFileDialogPaths(DlgName: string): string;
begin
  Result:=Options.LastFileDialogPaths.Values['DlgName'];
end;

procedure TApplicationDM.SetLastFileDialogPaths(DlgName: string; Path: string);
begin
  Options.LastFileDialogPaths.Values['DlgName']:=Path;
end;

// -------------------------------------------------
// MYXOptions

constructor TMYXAdminOptions.Create;
begin
  LastFileDialogPaths:=TStringList.Create;

  inherited Create;
end;

destructor TMYXAdminOptions.Destroy;
begin
  inherited Destroy;

  LastFileDialogPaths.Free;
end;

procedure TMYXAdminOptions.LoadOptions;
var POptions: PMYX_APPLICATION_OPTIONS;
  Options: TMYX_APPLICATION_OPTIONS;
  i, j: integer;
  error: MYX_LIB_ERROR;
  OptionGroupName, OptionName, OptionValue: string;
  ExePath: string;
begin
  ExePath:=ExtractFilePath(Application.ExeName);

  StartSection:=1;
  ShowOnlyServiceSections:=False;

  if(DirectoryExists(GetProgramFilesDir+'mysql'))then
    MySQLInstallPath:=GetProgramFilesDir+'mysql\'
  else if(DirectoryExists('c:\mysql'))then
    MySQLInstallPath:='c:\mysql\'
  else if(DirectoryExists('d:\mysql'))then
    MySQLInstallPath:='d:\mysql\'
  else
    MySQLInstallPath:=IncludeTrailingPathDelimiter(
      ExtractFilePath(Application.ExeName));

  MySQLVersion:=5.0;


  ShowTipOfDay:=True;

  //Update from 1.0.1a to 1.0.2a
  if(FileExists(MYXCommonOptions.UserXMLDir+'mysqlx_options.xml'))then
    RenameFile(MYXCommonOptions.UserXMLDir+'mysqlx_options.xml',
      MYXCommonOptions.UserXMLDir+'mysqlx_admin_options.xml');

  //------------------------------------------------------------
  //Initialize Values
  ShowTipOfDay:=False;
  StoreWindowsPositions:=False;

  SectionSidebarWidth:=185;
  SectionSidebarHidden:=False;
  ShowUserGlobalPrivileges:=False;
  ShowUserTableColumnPrivileges:=False;

  UsePeakLevel:=True;
  ResetPeakLevel:=True;
  PeakLevelResetTicks:=30;

  LastFileDialogPaths.Text:='';

  //------------------------------------------------------------
  //Read Options file
  if(FileExists(MYXCommonOptions.UserXMLDir+'mysqlx_admin_options.xml'))then
  begin
    POptions:=myx_get_application_options(
      PChar(AnsiToUtf8(MYXCommonOptions.UserXMLDir+'mysqlx_admin_options.xml')),
      @error);
    try
      if(error<>MYX_NO_ERROR)then
      begin
        ShowModalDialog('XML Error',
          'Error while loading Options file '''+
          MYXCommonOptions.UserXMLDir+'mysqlx_admin_options.xml'+''''#13#10+
          'Error Nr.:'+IntToStr(Ord(error)),
          myx_mtError);
      end
      else
      begin
        Options:=TMYX_APPLICATION_OPTIONS.Create(POptions);
        try
          for i:=0 to Options.option_groups.Count-1 do
            for j:=0 to Options.option_groups[i].name_value_pairs.Count-1 do
            begin
              OptionGroupName:=Options.option_groups[i].name;
              OptionName:=Options.option_groups[i].name_value_pairs[j].name;
              OptionValue:=Options.option_groups[i].name_value_pairs[j].value;

              if(CompareText(OptionGroupName, 'General')=0)then
              begin
                if(CompareText(OptionName, 'ShowTipOfDay')=0)then
                  ShowTipOfDay:=CheckBoolStr(OptionValue)
                else
                if(CompareText(OptionName, 'StoreWindowsPositions')=0)then
                  StoreWindowsPositions:=CheckBoolStr(OptionValue)
                else
                if(CompareText(OptionName, 'LastFileDialogPaths')=0)then
                  LastFileDialogPaths.Text:=AnsiReplaceText(OptionValue,
                    '', #13#10)
                else
                  ;
              end
              else if(CompareText(OptionGroupName, 'GUISetup')=0)then
              begin
                if(CompareText(OptionName, 'SectionSidebarWidth')=0)then
                  SectionSidebarWidth:=StrToIntDef(OptionValue, 185)
                else if(CompareText(OptionName, 'SectionSidebarHidden')=0)then
                  SectionSidebarHidden:=(StrToIntDef(OptionValue, 0)=1);
              end
              else if(CompareText(OptionGroupName, 'AdminUserManagement')=0)then
              begin
                if(CompareText(OptionName, 'ShowUserGlobalPrivileges')=0)then
                  ShowUserGlobalPrivileges:=(StrToIntDef(OptionValue, 0)=1)
                else if(CompareText(OptionName, 'ShowUserTableColumnPrivileges')=0)then
                  ShowUserTableColumnPrivileges:=(StrToIntDef(OptionValue, 0)=1);
              end
              else if(CompareText(OptionGroupName, 'AdminHealthGraphs')=0)then
              begin
                if(CompareText(OptionName, 'UsePeakLevel')=0)then
                  UsePeakLevel:=(StrToIntDef(OptionValue, 1)=1)
                else if(CompareText(OptionName, 'ResetPeakLevel')=0)then
                  ResetPeakLevel:=(StrToIntDef(OptionValue, 1)=1)
                else if(CompareText(OptionName, 'PeakLevelResetTicks')=0)then
                  PeakLevelResetTicks:=StrToIntDef(OptionValue, 30);
              end;


            end;
        finally
          Options.Free;
        end;
      end;
    finally
      //Free Application Options
      myx_free_application_options(POptions);
    end;
  end;
end;

procedure TMYXAdminOptions.StoreOptions;
var Options: TMYX_APPLICATION_OPTIONS;
  OptionGroup: TMYX_OPTION_GROUP;
  ExePath: string;
begin
  ExePath:=ExtractFilePath(Application.ExeName);

  //Create Application Options
  Options:=TMYX_APPLICATION_OPTIONS.create;
  try
    //-----------------------------------------------
    //Create Option Group
    OptionGroup:=TMYX_OPTION_GROUP.create('General');
    Options.option_groups.Add(OptionGroup);

    AddParam(OptionGroup, 'LastFileDialogPaths',
      AnsiReplaceText(LastFileDialogPaths.Text, #13#10, ''));

    //-----------------------------------------------
    //Create Option Group
    OptionGroup:=TMYX_OPTION_GROUP.create('GUISetup');
    Options.option_groups.Add(OptionGroup);

    AddParam(OptionGroup, 'SectionSidebarWidth',
      IntToStr(SectionSidebarWidth));
    AddParam(OptionGroup, 'SectionSidebarHidden',
      IntToStr(Ord(SectionSidebarHidden)));

    //-----------------------------------------------
    //Create Option Group
    OptionGroup:=TMYX_OPTION_GROUP.create('AdminUserManagement');
    Options.option_groups.Add(OptionGroup);

    AddParam(OptionGroup, 'ShowUserGlobalPrivileges',
      IntToStr(Ord(ShowUserGlobalPrivileges)));
    AddParam(OptionGroup, 'ShowUserTableColumnPrivileges',
      IntToStr(Ord(ShowUserTableColumnPrivileges)));


    //-----------------------------------------------
    //Create Option Group
    OptionGroup:=TMYX_OPTION_GROUP.create('AdminHealthGraphs');
    Options.option_groups.Add(OptionGroup);

    AddParam(OptionGroup, 'UsePeakLevel',
      IntToStr(Ord(UsePeakLevel)));
    AddParam(OptionGroup, 'ResetPeakLevel',
      IntToStr(Ord(ResetPeakLevel)));
    AddParam(OptionGroup, 'PeakLevelResetTicks',
      IntToStr(PeakLevelResetTicks));


    //-----------------------------------------------
    //Store Options to file
    myx_store_application_options(Options.get_record_pointer,
      PChar(AnsiToUtf8(MYXCommonOptions.UserXMLDir+'mysqlx_admin_options.xml')));
  finally
    Options.Free;
  end;
end;

end.
