/* Copyright (C) 2003 MySQL AB

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */


#include "myx_library.h"

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

#include <assert.h>
#include "myx_xml_aux_functions.h"

/*
 * Forward declarations
 */
static void get_name_value_pairs(xmlNodePtr node, MYX_OPTION_GROUP *option_group);


/*
 * Public functions
 */


/* loads application options from an XML file  */
MYX_APPLICATION_OPTIONS* myx_get_application_options(char *filename, MYX_LIB_ERROR *error_code)
{
    xmlDocPtr doc;
    xmlNodePtr root,cur;
    MYX_APPLICATION_OPTIONS *app_options;
    int i;

    *error_code= MYX_NO_ERROR;

    if (! file_exists(filename) )
    {
        *error_code = MYX_ERROR_CANT_OPEN_FILE;
        return NULL;
    }

    doc = xmlParseFile(filename);
    
    if (doc == NULL ) {
        *error_code = MYX_XML_PARSE_ERROR;
        return NULL;
    }
    
    root = xmlDocGetRootElement(doc);
    
    if (root == NULL) {
        *error_code = MYX_XML_EMPTY_DOCUMENT;
        xmlFreeDoc(doc);
        return NULL;
    }
    
    if (xmlStrcmp(root->name, (const xmlChar *) "application_options")) {
        *error_code = MYX_XML_NO_VALID_DOCUMENT;
        xmlFreeDoc(doc);
        return NULL;
    }
  

    app_options = malloc(sizeof(MYX_APPLICATION_OPTIONS));
    
 
    /*count the number of group_elements*/
    app_options->option_groups_num= 0;
    cur = root->children;
    while (cur != NULL) 
    {        
        if (! xmlStrcmp(cur->name, "group"))
            app_options->option_groups_num++;

        cur = cur->next;
    }

    if (app_options->option_groups_num > 0)
        app_options->option_groups = malloc(sizeof(MYX_OPTION_GROUP) * app_options->option_groups_num);
    else
        app_options->option_groups = NULL;

    /*read each group-element together with its name_value_pairs*/
    i=0;
    cur = root->children;
    while (cur != NULL) 
    {
        if (xmlStrcmp(cur->name, "group"))
        {
            cur = cur->next;
            continue;
        }

        app_options->option_groups[i].name = xmlGetProp(cur, "name");;
        get_name_value_pairs(cur, app_options->option_groups+i);

        i++;
        cur = cur->next;
    }


    xmlFreeDoc(doc);
    
    return app_options;
}

/* stores application option in an XML file */
/* returns 0 on success*/
int myx_store_application_options(MYX_APPLICATION_OPTIONS *options, char *filename)
{
    xmlDocPtr doc;
    xmlNodePtr root_node, group_node, name_value_node;
    unsigned int i,j;


    if ((options == NULL) || (filename == NULL)) return -1;

    doc = xmlNewDoc("1.0");

    root_node = doc->children = xmlNewDocRawNode(doc,NULL, "application_options", NULL);  

    for (i=0; i < options->option_groups_num; i++)
    {
        group_node = xmlNewTextChild(root_node, NULL, "group", NULL);
        xmlNewProp(group_node, "name", options->option_groups[i].name);

        for (j=0; j < options->option_groups[i].name_value_pairs_num; j++)
        {
            name_value_node = xmlNewTextChild(group_node, NULL, "property", NULL);

            xmlNewProp(name_value_node, "name", options->option_groups[i].name_value_pairs[j].name);
            xmlNewProp(name_value_node, "value", options->option_groups[i].name_value_pairs[j].value);
        }
        
    }


    if (xmlSaveFormatFile(filename, doc, 1) == -1)
    {
        xmlFreeDoc(doc);
        return -1;
    }

    xmlFreeDoc(doc);
    return 0;
}

MYX_PUBLIC_FUNC int myx_free_application_options(MYX_APPLICATION_OPTIONS *options)
{
  //TODO!!!
  return 0;
}


/*
 * Private functions
 */



/* Reads in the name-value pairs for a given group
 */
static void get_name_value_pairs(xmlNodePtr node, MYX_OPTION_GROUP *option_group)
{
    xmlNodePtr cur;
    int i;   

    /*count the number of name-value pairs*/
    option_group->name_value_pairs_num = 0;
    cur = node->children;
    while (cur)
    {
        if (!xmlStrcmp(cur->name, "property"))
            option_group->name_value_pairs_num++;

        cur = cur->next;
    }

    if (option_group->name_value_pairs_num > 0)
        option_group->name_value_pairs = malloc(sizeof(MYX_NAME_VALUE_PAIR) * option_group->name_value_pairs_num);
    else
        option_group->name_value_pairs = NULL;

    /*read in the name-value pairs*/
    i=0;
    cur = node->children;
    while (cur)
    {
        if (xmlStrcmp(cur->name, "property") ) /*ignore wrong elements*/
        {
            cur = cur->next;
            continue;
        }

        option_group->name_value_pairs[i].name = xmlGetProp(cur, "name");
        option_group->name_value_pairs[i].value = xmlGetProp(cur, "value");

        i++;
        cur = cur->next;
    }
}

