/* Copyright (C) 2003 MySQL AB

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */


#include <myx_library.h>

#include <glib.h>

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <libxml/xpath.h>

#include <assert.h>
#include <myx_xml_aux_functions.h>
#include <string.h>

/*static variables*/
static xmlXPathContextPtr context; /* used by all XPath operations*/


/* forward declarations */
static xmlChar *obscure(const xmlChar* str);
static xmlChar *retrieve_password(xmlChar *password, enum myx_password_storage_type storage_type);
static xmlChar* save_password(xmlChar *password, enum myx_password_storage_type storage_type);
static void parse_user_connection(xmlDocPtr doc, xmlNodePtr user_conn , MYX_USER_CONNECTION *out, enum myx_password_storage_type pwd_storage_type);
static void free_user_connection_content(MYX_USER_CONNECTION *uc);

/*
 * Public functions
 */


MYX_USER_CONNECTIONS* myx_load_user_connections(const char *filename, MYX_LIB_ERROR *error_code)
{
    MYX_USER_CONNECTIONS *user_conns;   
    xmlDocPtr doc;
    xmlNodePtr root;
    int count, i;
    xmlNodeSetPtr node_set;
    xmlXPathObjectPtr result;
    enum myx_password_storage_type pwd_storage_type;

    *error_code= MYX_NO_ERROR;

    if (! file_exists(filename) )
    {
        user_conns= malloc(sizeof(MYX_USER_CONNECTIONS));
        user_conns->last_connection= -1;
        user_conns->user_connections= NULL;
        user_conns->user_connections_num= 0;

        return user_conns;
    }


    doc = xmlParseFile(filename);
    
    if (doc == NULL ) {
        *error_code = MYX_XML_PARSE_ERROR;
        return NULL;
    }
    
    root = xmlDocGetRootElement(doc);
    
    if (root == NULL) {
        *error_code = MYX_XML_EMPTY_DOCUMENT;
        xmlFreeDoc(doc);
        return NULL;
    }
    
    if (xmlStrcmp(root->name, (const xmlChar *) "user_connections")) {
        *error_code = MYX_XML_NO_VALID_DOCUMENT;
        xmlFreeDoc(doc);
        return NULL;
    }

    context = xmlXPathNewContext(doc); /*create a new context for use with the XPath functions*/
    
    user_conns=malloc(sizeof(MYX_USER_CONNECTIONS));

    /*retrieve the last_connection*/
    result = xmlXPathEval("/user_connections/last_connection", context);
    node_set = result->nodesetval;         
    assert(node_set->nodeNr == 1);
    user_conns->last_connection = atoi_and_free( xmlNodeListGetString(doc, node_set->nodeTab[0]->children, 1) );
    xmlXPathFreeObject(result); 

    /*retrieve the password_storage_type*/
    result = xmlXPathEval("/user_connections/password_storage_type", context);
    node_set = result->nodesetval;
    assert(node_set->nodeNr == 1);
    pwd_storage_type = atoi_and_free( xmlNodeListGetString(doc, node_set->nodeTab[0]->children, 1) );
    xmlXPathFreeObject(result); 

    /*retrieve the user_connections*/
    result = xmlXPathEval("/user_connections/user_connection", context);
    node_set = result->nodesetval;


    if (xmlXPathNodeSetIsEmpty(node_set))
    {
        count = 0;
        user_conns->user_connections_num = 0;
        user_conns->user_connections = NULL;
    } else {
        count = node_set->nodeNr;
        user_conns->user_connections_num = count;
        user_conns->user_connections = malloc(sizeof(MYX_USER_CONNECTION) * count);
    }
    
    for (i=0; i< count; i++) 
    {
        parse_user_connection(doc,node_set->nodeTab[i], user_conns->user_connections+i, pwd_storage_type);
    }
    
    xmlXPathFreeObject(result); 

    xmlXPathFreeContext(context);
    xmlFreeDoc(doc);
    
    return user_conns;
}

int myx_store_user_connections(MYX_USER_CONNECTIONS* user_connections, enum myx_password_storage_type pwd_storage_type, const char *filename)
{
    xmlDocPtr doc;
    xmlNodePtr root_node, node, user_conn_node;
    int i,j;
    MYX_USER_CONNECTION *user_conn;

    if (user_connections == NULL) return -1;

    doc = xmlNewDoc("1.0");

    root_node = doc->children = xmlNewDocRawNode(doc,NULL, "user_connections", NULL);  

    xmlNewTextChild(root_node, NULL, "last_connection", int_to_str(user_connections->last_connection));
    xmlNewTextChild(root_node, NULL, "password_storage_type", int_to_str(pwd_storage_type));
   
#define STRNULL(s) (s)?(s):""
    
    for (i=0; i < (int)user_connections->user_connections_num; i++)
    {
        user_conn = user_connections->user_connections+i;
        user_conn_node = xmlNewTextChild(root_node, NULL, "user_connection", NULL);

        xmlNewTextChild(user_conn_node, NULL, "connection_name", STRNULL(user_conn->connection_name));
        xmlNewTextChild(user_conn_node, NULL, "username",user_conn->username);        
        xmlNewTextChild(user_conn_node, NULL, "hostname", user_conn->hostname);
        NewTextChild_int_content(user_conn_node, NULL, "port", user_conn->port);
        xmlNewTextChild(user_conn_node, NULL, "schema", user_conn->schema);

        node = xmlNewTextChild(user_conn_node, NULL, "advanced_options", NULL);
        for (j=0; j < (int)user_conn->advanced_options_num; j++)
        {
            xmlNewTextChild(node, NULL, "advanced_option",user_conn->advanced_options[j]);
        }

        xmlNewTextChild(user_conn_node, NULL, "storage_path", STRNULL(user_conn->storage_path));
        xmlNewTextChild(user_conn_node, NULL, "notes", STRNULL(user_conn->notes));

        NewTextChild_int_content(user_conn_node, NULL, "connection_type", user_conn->connection_type);
        NewTextChild_int_content(user_conn_node, NULL, "storage_type", user_conn->storage_type);
        xmlNewTextChild(user_conn_node, NULL, "password", save_password(user_conn->password, pwd_storage_type));
    }
#undef STRNULL

    if (xmlSaveFormatFile(filename, doc, 1) == -1)
    {
        xmlFreeDoc(doc);
        return -1;
    }

    xmlFreeDoc(doc);
    return 0;
}


int myx_free_user_connections(MYX_USER_CONNECTIONS *user_connections)
{
  unsigned int i;
  if (user_connections != NULL)
  {
    for (i=0; i < user_connections->user_connections_num; i++)
    {
      free_user_connection_content(user_connections->user_connections+i);
    }
    free(user_connections);
  }
  return 0;
}


int myx_add_user_connection(MYX_USER_CONNECTIONS **user_connections,
                            MYX_USER_CONNECTION *new_connection)
{
  unsigned int i;
  int found= 0;

  if (*user_connections)
  {
    for (i= 0; i < (*user_connections)->user_connections_num; i++)
    {
      if ((*user_connections)->user_connections[i].connection_name &&
          strcmp2((*user_connections)->user_connections[i].connection_name,
                 new_connection->connection_name)==0)
      {
        found= 1;
        break;
      }
    }
  }
  else
  {
    *user_connections= (MYX_USER_CONNECTIONS*)g_malloc0(sizeof(MYX_USER_CONNECTIONS));
  }
  if (!found)
  {
    MYX_USER_CONNECTION *ptr;

    (*user_connections)->user_connections_num++;
    (*user_connections)->user_connections= (MYX_USER_CONNECTION*)
      g_realloc((*user_connections)->user_connections, 
              sizeof(MYX_USER_CONNECTION)*(*user_connections)->user_connections_num);
    ptr= (*user_connections)->user_connections + ((*user_connections)->user_connections_num-1);

#define DUPN(s) (s != NULL ? xmlMemoryStrdup(s) : NULL)
    
    ptr->connection_name= DUPN(new_connection->connection_name);
    ptr->username= xmlMemoryStrdup(new_connection->username);
    ptr->password= xmlMemoryStrdup(new_connection->password);
    ptr->hostname= xmlMemoryStrdup(new_connection->hostname);
    ptr->port= new_connection->port;
    ptr->schema= xmlMemoryStrdup(new_connection->schema);
    ptr->advanced_options_num= new_connection->advanced_options_num;
    if (ptr->advanced_options_num > 0)
    {
      ptr->advanced_options= g_malloc(sizeof(char*)*ptr->advanced_options_num);
      for (i= 0; i < ptr->advanced_options_num; i++)
      {
        ptr->advanced_options[i]= xmlMemoryStrdup(new_connection->advanced_options[i]);
      }
    }
    else
      ptr->advanced_options= NULL;
    ptr->storage_path= DUPN(new_connection->storage_path);
    ptr->notes= DUPN(new_connection->notes);
    ptr->connection_type= new_connection->connection_type;
    ptr->storage_type= new_connection->storage_type;

#undef DUPN

    return (*user_connections)->user_connections_num-1;
  }

  return i;
}


/*
 * Private functions
 */

static void free_user_connection_content(MYX_USER_CONNECTION *uc)
{
  unsigned int i;

  xmlFree(uc->connection_name);
  xmlFree(uc->username);
  xmlFree(uc->password);
  xmlFree(uc->hostname);
  xmlFree(uc->schema);

  for (i=0; i < uc->advanced_options_num; i++)
  {
    xmlFree(uc->advanced_options[i]);
  }
  xmlFree(uc->storage_path);
  xmlFree(uc->notes);

  free(uc->advanced_options);
}


/* This function is used to encode binary strings into a string of hex digits */
static xmlChar *hex_encode(const xmlChar *str)
{
    unsigned int i;
    unsigned int len= (unsigned int)strlen(str)+1;
    xmlChar *result= malloc(len*2+1);

    *result= 0;
    for (i= 0; i < len; i++)
    {
	    sprintf(result+i*2, "%x", str[i]);   
    }
    result[i*2+1]= 0;

    return result;
}


static xmlChar *hex_decode(const xmlChar *str)
{
    unsigned int i;
    unsigned int len= (unsigned int)strlen(str)+1;
    xmlChar *result= malloc(len/2+2);

    *result= 0;
    for (i= 0; i < len/2; i++)
    {
	    unsigned int num;
	    sscanf(str+i*2, "%2x", &num);
	    result[i]= num;
    }
    result[i]= 0;
    return result;
}



/* Pseudo-encryption */
/* This function is used for encrypting as well as decrypting*/
static xmlChar *obscure(const xmlChar* str)
{
    int len,i;
    xmlChar *result;
    
    if(str)
    {
      len = (long)strlen(str);
      result = malloc(len+1);

      for (i=0; i < len; i++)
      {
          result[i] = ~str[i];
      }
      result[i] = '\0';

      return result;
    }
    else
    {
      return 0;
    }
}

/* always returns the password*/
static xmlChar *retrieve_password(xmlChar *password, enum myx_password_storage_type storage_type)
{
    xmlChar *plaintext_password, *tmp, *tmp2;


    switch (storage_type) {

    case MYX_PASSWORD_NOT_STORED:
        plaintext_password= NULL;
        break;
    case MYX_PASSWORD_PLAINTEXT:
        plaintext_password= password;
        break;
    case MYX_PASSWORD_OBSCURED:
	    if (password)
	    {
	        tmp= hex_decode(password);
          xmlFree(password);
          tmp2= obscure(tmp);
          plaintext_password= xmlStrdup(tmp2);
	        free(tmp);
          free(tmp2);
	    }
	    else
	    {
	        plaintext_password= NULL;
	    }
      break;
    case MYX_PASSWORD_OS_SPECIFIC:
    default:
//#warning TODO here
        plaintext_password= password;
        break;
    }

    return plaintext_password;
}

static xmlChar* save_password(xmlChar *password, enum myx_password_storage_type storage_type)
{
    xmlChar *encrypted_password, *tmp;

    switch (storage_type) {

    case MYX_PASSWORD_NOT_STORED:
        encrypted_password = NULL;
        break;
    case MYX_PASSWORD_PLAINTEXT:
        encrypted_password = password;
        break;
    case MYX_PASSWORD_OBSCURED:
	if (password)
	{
 	    tmp= obscure(password);
            encrypted_password = hex_encode(tmp);
	    free(tmp);
	}
	else
	{
	    encrypted_password= NULL;
	}
        break;
    case MYX_PASSWORD_OS_SPECIFIC:
    default:
        encrypted_password = password;
        break;
    }

    return encrypted_password;
}

/*used for parsing a user_connection of the xml-dom-tree*/
static void parse_user_connection(xmlDocPtr doc, xmlNodePtr user_conn , MYX_USER_CONNECTION *out, enum myx_password_storage_type pwd_storage_type)
{
    xmlNodePtr cur;
    int i;
    xmlXPathObjectPtr result;
    xmlNodeSetPtr node_set;

    assert( !xmlStrcmp(user_conn->name, "user_connection") );

    memset(out, 0, sizeof(MYX_USER_CONNECTION));
    
    cur = user_conn->children;
    while (cur != NULL) 
    {
        if ( !xmlStrcmp(cur->name, "connection_name") )
            out->connection_name = xmlNodeListGetString(doc, cur->children, 1);

        if ( !xmlStrcmp(cur->name, "username") )
            out->username = xmlNodeListGetString(doc, cur->children, 1);

        if ( !xmlStrcmp(cur->name, "hostname") )
            out->hostname = xmlNodeListGetString(doc, cur->children, 1);

        if ( !xmlStrcmp(cur->name, "port") )
            out->port = atoi_and_free(xmlNodeListGetString(doc, cur->children, 1));

        if ( !xmlStrcmp(cur->name, "schema") )
            out->schema = xmlNodeListGetString(doc, cur->children, 1);

        if ( !xmlStrcmp(cur->name, "connection_type") )
            out->connection_type = atoi_and_free(xmlNodeListGetString(doc, cur->children, 1));

        if ( !xmlStrcmp(cur->name, "storage_type") )
            out->storage_type = atoi_and_free(xmlNodeListGetString(doc, cur->children, 1));

        if ( !xmlStrcmp(cur->name, "password") )
            out->password = retrieve_password(xmlNodeListGetString(doc, cur->children, 1), pwd_storage_type);

        if ( !xmlStrcmp(cur->name, "storage_path") )
            out->storage_path = xmlNodeListGetString(doc, cur->children, 1);

        if ( !xmlStrcmp(cur->name, "notes") )
            out->notes = xmlNodeListGetString(doc, cur->children, 1);

        cur = cur->next;
    }

    context->node = user_conn;
    result = xmlXPathEval("advanced_options/advanced_option", context);
    node_set = result->nodesetval;

    if (xmlXPathNodeSetIsEmpty(node_set))
    {
        out->advanced_options_num = 0;
        out->advanced_options = NULL;
    } 
    else 
    {
        out->advanced_options_num = node_set->nodeNr;
        out->advanced_options = malloc(sizeof(char *) * node_set->nodeNr);
    }
    
    for (i=0; i< (int)out->advanced_options_num; i++) 
    {
        out->advanced_options[i] = xmlNodeListGetString(doc, node_set->nodeTab[i]->children, 1);
    }
}

