unit Options;

interface

uses
  Classes, Forms, SysUtils, Messages, ComCtrls,
  myx_public_interface, AuxFuncs;

const
  WM_OptionsChanged = WM_USER+100;

type
  TDoPageContentChanged = procedure (Sender: TObject) of object;

  TMyxOptions = Class(TObject)
    constructor Create;
    destructor Destroy; override;

    procedure LoadOptions; virtual; abstract;
    procedure StoreOptions; virtual; abstract;

    function CheckBoolStr(s: string): Boolean;
    procedure AddParam(OptionGroup: TMYX_OPTION_GROUP;
      name: string; value: string);
  end;

  TMYXCommonOptions = Class(TMyxOptions)
    procedure LoadOptions; override;
    procedure StoreOptions; override;
  public
    AvailableLanguages: string;
    AvailableLanguageCodes: string;
    AvailablePasswordStorageTypes: string;

    Language: string;

    XPStyleEnabled: Boolean;

    FileDialogPaths: TStringList;

    XMLDir, UserXMLDir: string;

    PasswordStorageType: MYX_PASSWORD_STORAGE_TYPE;
  end;

  TApplicationOptionsForm = class(TForm)
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure SetControls(PageNr: integer); virtual; abstract;
    procedure ApplyChanges(PageNr: integer); virtual; abstract;
  public
    { Public declarations }
    OptionsImgNames: TStringList;
    DockOptionPageControl: TPageControl;
    DoPageContentChanged: TDoPageContentChanged;
  end;

var
  MYXCommonOptions: TMYXCommonOptions;

implementation

uses ApplicationDataModule;

constructor TMyxOptions.Create;
begin
  inherited Create;

  LoadOptions;
end;

destructor TMyxOptions.Destroy;
begin
  StoreOptions;

  inherited Destroy;
end;

// Help functions

function TMyxOptions.CheckBoolStr(s: string): Boolean;
begin
  Result:=(StrToIntDef(s, 1)=1);
end;

procedure TMyxOptions.AddParam(OptionGroup: TMYX_OPTION_GROUP;
  name: string; value: string);
begin
  OptionGroup.name_value_pairs.Add(
    TMYX_NAME_VALUE_PAIR.create(name, value));
end;


// -------------------------------------------------
// TMYXCommonOptions

procedure TMYXCommonOptions.LoadOptions;
var POptions: PMYX_APPLICATION_OPTIONS;
  Options: TMYX_APPLICATION_OPTIONS;
  i, j: integer;
  error: MYX_LIB_ERROR;
  OptionGroupName, OptionName, OptionValue: string;
  ExePath: string;
begin
  ExePath:=ExtractFilePath(Application.ExeName);

  //Detect if XPStyle is enabled
  XPStyleEnabled:=getXPStyleEnabled;

  AvailableLanguages:='English';//#13#10'German';
  AvailableLanguageCodes:='en';//#13#10'de';
  AvailablePasswordStorageTypes:='Plaintext'#13#10'Obscured';//#13#10'OS Specific';

  Language:='en';

  XMLDir:=ExtractFilePath(Application.ExeName)+'XML\';
  UserXMLDir:=GetApplicationDataDir+'MySQL\';
  ForceDirectories(UserXMLDir);

  //------------------------------------------------------------
  //Initialize Values
  Language:='en';

  //Initialize Default Values
  PasswordStorageType:=MYX_PASSWORD_NOT_STORED;


  //------------------------------------------------------------
  //Read Options file
  if(FileExists(UserXMLDir+'mysqlx_common_options.xml'))then
  begin
    POptions:=myx_get_application_options(
      PChar(AnsiToUtf8(UserXMLDir+'mysqlx_common_options.xml')),
      @error);
    try
      if(error<>MYX_NO_ERROR)then
      begin
        ShowModalDialog('XML Error',
          'Error while loading Options file '''+
          UserXMLDir+'mysqlx_common_options.xml'+''''#13#10+
          'Error Nr.:'+IntToStr(Ord(error)),
          myx_mtError);
      end
      else
      begin
        Options:=TMYX_APPLICATION_OPTIONS.Create(POptions);
        try
          for i:=0 to Options.option_groups.Count-1 do
            for j:=0 to Options.option_groups[i].name_value_pairs.Count-1 do
            begin
              OptionGroupName:=Options.option_groups[i].name;
              OptionName:=Options.option_groups[i].name_value_pairs[j].name;
              OptionValue:=Options.option_groups[i].name_value_pairs[j].value;

              if(CompareText(OptionGroupName, 'General')=0)then
              begin
                if(CompareText(OptionName, 'Language')=0)then
                  Language:=OptionValue
                else if(CompareText(OptionName, 'PasswordStorageType')=0)then
                  PasswordStorageType:=MYX_PASSWORD_STORAGE_TYPE(StrToIntDef(OptionValue, 1))
                else
                  ;

              end;
            end;
        finally
          Options.Free;
        end;
      end;
    finally
      //Free Application Options
      myx_free_application_options(POptions);
    end;
  end;

end;

procedure TMYXCommonOptions.StoreOptions;
var Options: TMYX_APPLICATION_OPTIONS;
  OptionGroup: TMYX_OPTION_GROUP;
  ExePath: string;
begin
  ExePath:=ExtractFilePath(Application.ExeName);

  //Create Application Options
  Options:=TMYX_APPLICATION_OPTIONS.create;
  try
    //-----------------------------------------------
    //Create Option Group
    OptionGroup:=TMYX_OPTION_GROUP.create('General');
    Options.option_groups.Add(OptionGroup);

    //Add Values
    AddParam(OptionGroup, 'Language', Language);

    //Add Values
    AddParam(OptionGroup, 'PasswordStorageType',
      IntToStr(Ord(PasswordStorageType)));

    //-----------------------------------------------
    //Store Options to file
    myx_store_application_options(Options.get_record_pointer,
      PChar(AnsiToUtf8(UserXMLDir+'mysqlx_common_options.xml')));
  finally
    Options.Free;
  end;
end;

constructor TApplicationOptionsForm.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  OptionsImgNames:=TStringList.Create;
  DockOptionPageControl:=nil;
  DoPageContentChanged:=nil;
end;

destructor TApplicationOptionsForm.Destroy;
begin
  OptionsImgNames.Free;

  inherited Destroy;
end;


end.
